###############################################################################
#
# DC.pm
#
# Common utility package for Sega GT CGI programs
#
#
#
###############################################################################

package DC;

use strict;

BEGIN {
	use Exporter	();
	use	CGI			();
	use CGI::Carp ('fatalsToBrowser');
	use DB;
	require "bigint.pl";

	use POSIX	qw(strftime ceil);
	use vars	qw($VERSION @ISA @EXPORT @EXPORT_OK %EXPORT_TAGS $AUTOLOAD);

	$VERSION = do {my @r = (2.00); sprintf "%d."."%02d"x$#r, @r;};

	@ISA = qw(CGI);
	@EXPORT = qw();
	%EXPORT_TAGS = ( );
	@EXPORT_OK = qw ( );
}


# ********************************************************
# Contains Smut filter, State names, and Country Names
# ********************************************************
	use vars (	'@smut', '%StateName', '%CountryName', 
				'%Flags', '@KanjiCodeText', '%DBaseTable',
				'%UserOff', '%WRankOff',
				'@Browsers', '@base64', '@dp_base64', '%accessors'
			);

require 'strings.pl';
require 'defines.pl';

############################################################################
# %accessors defines what AUTOLOAD can see
# for simple get/set accessors
# this map operation simply results in a hash that looks like:
# ( foo=>undef, bar=>undef, baz=>undef )
############################################################################
%accessors = map { $_ => undef }
				qw ( q );

############################################################################
# User Agent strings for the Rush 2049 Browser
############################################################################
@Browsers = (
			# US Version
			"Mozilla/3.0 (DreamKey/1.0; SEGA/JETSETRADIO)",

			# European Version
			"",

			# Japanese Version
			"",
		);

############################################################################
# Base 64 Decoding Table for VMU File Uploads
############################################################################
# encryption table
@base64    = ('A','B','C','D','E','F','G','H','I','J',
              'K','L','M','N','O','P','Q','R','S','T',
              'U','V','W','X','Y','Z','a','b','c','d',
              'e','f','g','h','i','j','k','l','m','n',
              'o','p','q','r','s','t','u','v','w','x',
              'y','z','0','1','2','3','4','5','6','7',
              '8','9','+','/','=');

@dp_base64 = ('A','Z','O','L','Y','N','d','n','E','T',
              'm','P','6','c','i','3','S','z','e','9',
              'I','y','X','B','h','D','g','f','Q','q',
              '7','l','5','b','a','t','M','4','r','p',
              'K','J','j','8','C','u','s','x','R','F',
              '+','k','2','V','0','w','U','G','o','1',
              'v','W','H','/','=');

# ********************************************************
#############################################################################
# new (<query object>)
#############################################################################
sub new {
	my $proto = shift;
	my $class = ref($proto) || $proto;

	my $query = shift;

	my $self = {};
	bless $self, $class;

	## at some point this should politely die, instead of die().
	(defined($query) && $self->q($query)) || die "No query provided!";

	return $self;
}

#####################################################################
# CalcTime()
# Input : Time value
# Output : Minutes, seconds, miliseconds
#####################################################################
sub CalcTime()	{
	my $self = shift;		# Pointer to CGI structure
	my $time = shift;

	my ($min, $sec, $msec, $str);

	$min = int($time / 60);
	$sec = $time - ($min * 60);
	$str = sprintf("%d:%02.3f", $min, $sec);

	return $str;
}
 
#############################################################################
# decode_hwid(<base 64'd id>)
# Decode Hardware ID
#############################################################################
sub decode_hwid {
	my $self = shift;		# Pointer to CGI structure
    my $srcdata = shift;
    my $hwid = "";

    # BASE64Decode
    my @val = DC::decode_base64($srcdata);
    for(my $i=0;$i<8;$i++) {
        $hwid .= sprintf "%02X",$val[$i];
    }

    return $hwid;
}

#############################################################################
# decode_base64(<base 64'd data stream>)
# returns an array derived from base64 decoding the data stream.
#
# Argument: Character lines encoded by BASE64.
# Back: Encoded bit character lines.
############################################################################
sub decode_base64 {
	my $self = shift;		# Pointer to CGI structure
    my $src = shift;
 
    my @bindata = ();
    my $bytes =0;
    my $tmpstr = "";
    my $bits ="";
    foreach my $c (split(//,$src)) {
        if($c eq '=') {
            last;
        }
        if($c =~ /[A-Z]/) {
            $bits .= substr(unpack("B*",chr(ord($c)-ord('A'))),-6,6);
        } elsif($c =~ /[a-z]/) {
            $bits .= substr(unpack("B*",chr(ord($c)-ord('a')+26)),-6,6);
        } elsif($c =~ /[0-9]/) {
            $bits .= substr(unpack("B*",chr(ord($c)-ord('0')+52)),-6,6);
        } elsif($c  eq '+') {
            $bits .= substr(unpack("B*",chr(62)),-6,6);
        } elsif($c  eq '/') {
            $bits .= substr(unpack("B*",chr(63)),-6,6);
        }

		## now stuff it per-byte into @bindata
        if(length($bits) >= 8) {
            $bindata[$bytes++] = ord(pack("B*",substr($bits,0,8)));
            $bits = substr($bits,8,length($bits)-8);
        }
    }

    return @bindata;
}

# ------------------- Not Used ------------------
#############################################################################
# encode_base64(<raw data stream>)
# returns an array derived from base64 decoding the data stream.
#
# Argument: Raw binary data.
# Back: Encoded bit character lines.
############################################################################
sub encode_base64 {
	my $self = shift;		# Pointer to CGI structure
    my $src = shift;
 
    my $retstr="";
    my $val=0;
    my $bitsnum=0;
    my $tmpval;

    foreach my $c (split(//,$src)) {
        $bitsnum+=8;
        $val=$val*256+pack("c",$c);
        while($bitsnum>=6) {
            $tmpval=($val>>($bitsnum-6))&63;
            $val=$val & ((1<<$bitsnum-6)-1);
            if($tmpval>=0 && $tmpval<26) {
                $retstr .= chr(ord('A')+$tmpval);
            } elsif($tmpval>=26 && $tmpval<52) {
                $retstr .= chr(ord('a')+$tmpval-26);
            } elsif($tmpval>=52 && $tmpval<62) {
                $retstr .= chr(ord('0')+$tmpval-52);
            } elsif($tmpval==62) {
                $retstr .= '+';
            } elsif($tmpval==63) {
                $retstr .= '/';
            }
            if((length($retstr)%73)==72) {
                $retstr .= "\n";
            }
        }
    }
    if($bitsnum>0) {
        $tmpval=($val<<(6-$bitsnum))&63;
        if($tmpval>=0 && $tmpval<26) {
            $retstr .= chr(ord('A')+$tmpval);
        } elsif($tmpval>=26 && $tmpval<52) {
            $retstr .= chr(ord('a')+$tmpval-26);
        } elsif($tmpval>=52 && $tmpval<62) {
            $retstr .= chr(ord('0')+$tmpval-52);
        } elsif($tmpval==62) {
            $retstr .= '+';
        } elsif($tmpval==63) {
            $retstr .= '/';
        }
        if((length($retstr)%73)==72) {
            $retstr .= "\n";
        }
    }
    $retstr .= "==\n";

    return $retstr;
}

#############################################################################
# decode_dp_base64(<DP base 64'd data stream>)
# returns an array derived from DP base64 decoding the data stream.
#
# Argument: Character lines encoded by DP_BASE64.
# Back: Encoded bit character lines.
############################################################################
sub decode_dp_base64 {
	my $self = shift;		# Pointer to CGI structure
	my $src = shift;

	my $bindata = ();
	my $i;

	my $count = 0;
	foreach my $c(split(//,$src))
	{
		if ($c eq '-') {
			last;
		}

		for ($i = 0; $i < 65; $i++)
		{
			if ($c eq $dp_base64[$i]) {
				$count++;
 				$bindata .= $base64[$i];
				if ($count == 77) {
					$count = 0;
					$bindata .= "\n";
				}
				last;
			}
		}
	}

	return $bindata;
}

#############################################################################
# read_multipart_post()
# Read the multipart post request and split into descrete blocks
#
# Argument: None.
# Back: Each component of the Multi-part type.
#############################################################################
sub read_multipart_post {
	my $self = shift;		# Pointer to CGI structure

	my ($formData, $size, @boundry, @returnData, %out, $tmpVar);

    if($ENV{'REQUEST_METHOD'} eq "GET") {
        $formData = $ENV{'QUERY_STRING'};
    }
    elsif($ENV{'REQUEST_METHOD'} eq "POST") {
        $size = $ENV{'CONTENT_LENGTH'};
        read(STDIN, $formData, $size);
    }
    else
    {
        $self->error("[DC.pm] Problem figuring out GET or POST\n");
    }

    @boundry = split(/\s/, $formData);

    @returnData = split(/$boundry[0]([-]{0,2})/, $formData);
    foreach $tmpVar (@returnData) {
        my $junk = "";
        my $key = "";
        my $value = "";

		if ($tmpVar ne "")
		{
			if ($tmpVar =~ /Content-Disposition: form-data; name=\"(.*)\"/)
			{
				$key = $1;
				($junk, $value) = split(/\".*\"/, $tmpVar, 2);
			}

			if ($key ne "")
			{
				$out{$key} = $value;
			}
		}
	}

    return %out;
}

#############################################################################
# read_vmu_upload()
# Read the VMU Upload Post request, and decode the VMU data
#
# Argument: None.
# Back: Returns an array to the VMU data.
#############################################################################
sub read_vmu_upload {
	my $self = shift;		# Pointer to CGI structure

    my @BINDATA = ();
	my ($key, $value, $tmpVar, @vmuData, %vmuInfo, $filename, $filesize);

	# Get the Post or Get request
	my %toPrint = DC::read_multipart_post();

	while(($key, $value) = each %toPrint)
	{
		if ($key eq "VMUDATA")
		{
			@vmuData = split(/&/, $value);

			foreach $tmpVar (@vmuData) {
				($key, $value) = split(/=/, $tmpVar, 2);
				$key =~ s/%([A-Fa-f0-9]{2})/pack("C",hex($1))/ge;
				$key =~ s/[\n\r]//g;
				$value =~ s/%([A-Fa-f0-9]{2})/pack("C",hex($1))/ge;
				$vmuInfo{$key} = $value;
			}

			while(($key, $value) = each %vmuInfo)
			{
				# get file name
				if ($key eq "filename") {
					$filename = $value;
				}

				# get file size
				if ($key eq "bl") {
					$filesize = $value;
				}

				# get binary data
				if ($key eq "tm")
				{
					my ($trim, $base64);
					$trim = substr($value, 15);
					$base64 = "";
					$base64 = DC::decode_dp_base64($trim);
					@BINDATA = DC::decode_base64($base64);
				}
			}
		}
	}

    return @BINDATA;
}

############################################################################
# make_vmidata(<filename on server>, <filesize in bytes>, <filename on VMU>)
# Creating VMI data.  Filename and VMU filename should only be Uppercase.
# The filename does not require the .VMS extension.  It is assumed.
# returns a VMI data structure for the specified VMU file.
#
# Argument: Filename on server, filesize in bytes, and VMU filename.
# Back: VMI data file.
############################################################################
sub make_vmidata {
	my $self = shift;		# Pointer to CGI structure

	my ($filename, $filesize, $vmufname) = @_;

	my ($i, @var, $FILE, @vmiheader);
	my @VMIDATA = ();

    @vmiheader = (
         "V", "M", "U", " ", "D", "O", "W", "N", "L", "O", "A", "D", " ", "F", "I", "L",
         "E","\0","\0","\0","\0","\0","\0","\0","\0","\0","\0","\0","\0","\0","\0","\0",
         "S", "E", "G", "A", " ", "E", "N", "T", "E", "R", "P", "R", "I", "S", "E", "S",
        "\0","\0","\0","\0","\0","\0","\0","\0","\0","\0","\0","\0","\0","\0","\0","\0"
        );

    for($i=0;$i<64;$i++) {
        $VMIDATA[$i+4] = ord($vmiheader[$i]);
    }

    my ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst) = localtime(time);
    $mon++;
    $VMIDATA[68]= ($year+1900 & 255);	# year (Low Byte)
    $VMIDATA[69]= ($year+1900 >> 8);	# year (High Byte)
    $VMIDATA[70]= $mon;			# month
    $VMIDATA[71]= $mday;		# day
    $VMIDATA[72]= $hour;		# hour
    $VMIDATA[73]= $min;			# minute
    $VMIDATA[74]= $sec;			# second
    $VMIDATA[75]= $wday;		# day of the week
    $VMIDATA[76]= 0;			# Version (inferior rank)
    $VMIDATA[77]= 1;			# Version (superior rank)
    $VMIDATA[78]= 1;			# # of files (inferior rank)
    $VMIDATA[79]= 0;			# # of files (superior rank)

    @var=split(//,$filename);
    for($i=0;$i<8;$i++) {
        $VMIDATA[$i+80] = ord($var[$i]);	# File names on server
    }

	$VMIDATA[0] = $VMIDATA[80] & 0x53;		# Checksum
	$VMIDATA[1] = $VMIDATA[81] & 0x45;		#
	$VMIDATA[2] = $VMIDATA[82] & 0x47;		#
	$VMIDATA[3] = $VMIDATA[83] & 0x41;		#

	$FILE = $vmufname;
    $FILE =~ y/a-z/A-Z/;

    @var=split(//,$FILE);
    for($i=0;$i<12;$i++) {
        $VMIDATA[$i+88] = ord($var[$i]);	# VMS file names
    }

	# Calculate Filesize
	$filesize = ($filesize / 512); # Convert to blocks

    $VMIDATA[100] = 0;			# Attributes
    $VMIDATA[101] = 0;			# Attributes
    $VMIDATA[102] = 0;			# Attributes
    $VMIDATA[103] = 0;			# Attributes
    $VMIDATA[104] = 0;			# File size
    $VMIDATA[105] = $filesize*2;	# File size
    $VMIDATA[106] = 0;			# File size
    $VMIDATA[107] = 0;			# File size

	my $vmidata = "";
	for($i=0;$i<108;$i++) {
        $vmidata .= pack("C",$VMIDATA[$i]);
    }

    return $vmidata;
}

#############################################################################
# make_key(<users id>)
# Take the users e-mail address and check to see if they exist in the
# database.  If they do not, then initialize them and give them a unique key
#
# Argument: E-mail address to search for
# Returns: Unique Key
#############################################################################
sub make_key {
	my $self = shift;		# Pointer to CGI structure
	my $userid = shift;
	my @row = ();
	my ($dbh, $errN, $errstr, $value);

	($dbh, $errstr) = DB::openOracleConnection();
	if( !$dbh )
	{
		$self->error("[make_key] Error connecting to database: $errstr");
	}

    ($errN, $errstr, @row) = DB::getRow_UserByUK( $dbh, $DBaseTable{'USER'}, $userid );
	if ($errN)
	{
		$self->error("[make_key] Fetching user key: $errN .... $errstr $userid");
	}

	# Setup the return value
	$value = $row[$self->UserOff('USERID')];
	if ($row[$self->UserOff('USERID')] eq "")
	{
		# User entry was not found
	   ($errN, $errstr, $value) = DB::insertRow_User( $dbh, $DBaseTable{'USER'}, $userid, "", $userid, "", "", "", "N", "N", 0, 0, 0, 0, time, time  );
		if ($errN)
		{
			$self->error("[make_key] Creating new user entry: $errN .... $errstr");
		}
	}

	DB::closeOracleConnection( $dbh );

	return $value;
}

#############################################################################
# get_key(<username>)
# Take the users login name and check to see if they exist in the
# database.  If they do then return the unique key.  Each username is unique.
#
# Argument: Username address to search for
# Returns: Unique Key ("" if not found)
#############################################################################
sub get_key {
	my $self = shift;		# Pointer to CGI structure
	my $username = shift;
	my $key = "";
	my ($dbh, $errN, $errstr);

	($dbh, $errstr) = DB::openOracleConnection();
	if( !$dbh )
	{
		$self->error("[get_key] Error connecting to database: $errstr");
	}

    ($errN, $errstr, $key) = DB::getUK_UserByName( $dbh, $DBaseTable{'USER'}, $username );
	if ($errN)
	{
		$self->error("[get_key] Fetching user key: $errN .... $errstr");
	}

	DB::closeOracleConnection( $dbh );

	return $key;
}

#############################################################################
# retBirthDaySec(<month>, <day>, <year>)
# Will convert birthday into 
#############################################################################
sub retBirthDaySec
{
	my($mm, $dd, $yy) = @_;
	my(@dmon) = ();
	my($ny, $i, $days);

	# Days in mon 0   1   2   3   4   5   6   7   8   9  10  11  12
	my (@dmon) = (0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31);

	$days = 0;
	for($i = 1900; $i < $yy; $i++)
	{
		$ny = ($i % 4 );
		if($ny == 0)
		{
			$days += 366;
		}
		else
		{
			$days += 365;
		}
	}
	$ny = ($yy % 4 );
	for($i = 1; $i < $mm; $i++)
	{
		$days += $dmon[$i];
		if($i == 2 && $ny == 0)
		{
			$days += 1;
		}
	}

	$days += $dd;
	return($days * 86400);
}

#############################################################################
# file_exists(<filename>)
# Read the VMU Upload Post request, and decode the VMU data
#
# Argument: filename to check.
# Back: Returns a non-zero value if file exists and has more than zero bytes.
#############################################################################
sub file_exists {
	my $self = shift;		# Pointer to CGI structure

	my $filename = shift;	# File to check to see if it exists

	my $result = 0;			# Used for handling errors
	my $check = 1;			# Check fails by default

	# Create the temporary image graphic if not already on server
	$result = open(TEST, $filename);
	if ($result == 0)
	{
		$check = 0;		# File doesn't exist so create it
	}
	else
	{
		# Make sure that the file size is non-zero
		$result = seek(TEST, 0, 2);
		if ($result == 0)
		{
			$check = 0;		# Seek failed so recreate file
		}
		else
		{
			# Return byte count, if no
			$result = tell(TEST);
			if ($result == 0)
			{
				$check = 0;
			}
		}
	}
	close (TEST);		# Done with the file handle

	return $check;
}

#############################################################################
# output_form(<template location>)
# takes a template and fills it in with the appropriate data from the query
#############################################################################
sub output_form {
	my $self = shift;		# Pointer to CGI structure
    my $filename = shift;

    if(open(TEMPLATE,$filename)) {
        print "Content-type: text/html\n\n";

        while(<TEMPLATE>) {
            while(s/([^<]*)(<[^<>]*>?)//) {
                print $1;
                my $tag = $2;
                if( $tag =~ /<!--Hidden:(.+)-->/ ) {
                    print "<INPUT TYPE=\"hidden\" NAME=\"$1\" VALUE=\"";
                    print $self->q->escapeHTML($self->q->param($1));
                    print "\">";
                } elsif( $tag =~ /<!--CB=(.+):(.+)-->/ ) {
                    if($self->q->param($1) eq $2) {
                        print "CHECKED"
                    }
                } elsif( $tag =~ /<!--SB=(.+):(.+)-->/ ) {
                    if($self->q->param($1) eq $2) {
                        print "SELECTED"
                    }
		} elsif( $tag =~ /<!--EMBLEMS=([0-9]+):([^,]+),(.+)-->/ ) {
		    print "<form action=\"";
		    if ($self->q->param('EMBLEMS') >= $1) {
			print "\"$2\">";
		    } else {
			print "\"$3\">";
		    }
                } elsif( $tag =~ /<!--(.+)-->/ ) {
					$self->q->autoEscape(0);
					print $self->q->param($1);
					$self->q->autoEscape(1);
                } else {
                    print $tag;
                }
            }
            print;
        }
        close(TEMPLATE);
    } else {
        $self->error("Failure to open \"$filename\".");
    }
}
 
############################################################################
# get_browser()
# returns a number from .
#
# Back: Return browser ID number. 0 = US, 1 = Europe, 2 = Japan
############################################################################
sub get_browser {
	my $self = shift;
	my $val = -1;

	my $browser = $self->q->user_agent();
	if (index($browser,$Browsers[0]) >= 0)
	{
		# US version of browser
        $val = 0;
	}
	elsif (index($browser,$Browsers[1]) >= 0)
	{
		# European version of browser
        $val = 1;
	}


	return $val;
}

############################################################################
# error(<error message>, <URL to return>)
# Error Message Display
#
# Argument: Error Message, and URL to return to
############################################################################
sub error {
	my $self = shift;		# Pointer to CGI structure
    my $mes = shift;
	my $URL = shift;

	unless (defined($URL)) {
		$URL = "index.html";
	}

    print <<EOF;
Content-type: text/html

<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.0 Transitional//EN">

<html>
<head>
	<title>Jet Grind Radio - Error</title>
<meta name=x-uirequest content=nohscroll>
</head>

<body bgcolor=#E6CE00 x-marginleft=0 TEXT="#660000" LINK="#B96800" VLINK="#B96800" ALINK="#B96800" marginheight=0 marginwidth=0 topmargin=0 leftmargin=0>



<center>
<TABLE WIDTH="550" BORDER=0 CELLPADDING=0 CELLSPACING=0>
	<tr>
		<TD valign=top width=550><img src="/images/title_error.gif" width=550 height=46 border=0 alt=""></td>
	</tr>
</table>
<TABLE WIDTH="550" BORDER=0 CELLPADDING=0 CELLSPACING=0>
	<tr>
		<TD width=7 valign=top><img src="/images/clear.gif" width=7 height=1 border=0 alt=""></td>
		<TD width=268 valign=top><br>
<b>--ERROR--</b><br><br>

$mes

<br><br><br><br>

<a href="http://jetgrindradio.web.dreamcast.com/$URL"><img src="/images/button_back.gif" width=200 height=30 border=1 alt="" align=right></a></td>
		<TD width=275 valign=top><img src="/images/graphic_login.gif" width=275 height=410 border=0 alt=""></td>
	</tr>
</table>
</body>

</html>
EOF
    exit;
}

#############################################################################
# smut_check(<string to be checked>)
# returns an string where words are filtered out, and also
# breaks any word longer than 62 characters.
#
# Argument: Raw string data.
# Back: Cleaned up string.
############################################################################
sub smut_check {
	my $self = shift;		# Pointer to CGI structure
    my $str = shift;
	my $censored = '#!@%#!@%#!@%#!@%#!@%#!@%#!@%#!@%#!@%';

	$str =~ s/</&lt;/gi;
	$str =~ s/>/&gt;/gi;

	# Smut checker
	for (my $loop = 0; ($smut[$loop] ne 'ENDOFLIST'); ++$loop)
	{
		$str =~ s/($smut[$loop])/substr($censored,0,length($1))/gie;
	}

	# Break up long words
	$str =~ s/(\b\S{25,}\b)/substr($1,0,25)." ".substr($1,25)/ge;

    return $str;
}

#############################################################################
# convert_katakana(<first character>, <second character>, <string>)
# returns a romanji string for the Katakana characters.
#
# Argument: First Katakana Character.
#           Second Katakana Character.
#			String to append converted text to
# Back: 0 if failed, 1 if succeeded with one character, 2 if took two characters
############################################################################
sub convert_katakana {
	my $self = shift;
	my $first = shift;
	my $second = shift;
	my $result = 0;
	my ($loop);

	if (($first >= 0x64) && ($first <= 0xB6))
	{
#		$self->error("First = ".$first.", Second = ".$second);	# For testing purposes
		# Name is in Kanji
		$loop = 0;
		while (($KanjiCodeText[$loop][0] != 0) && (!$result))
		{
			if ($first == $KanjiCodeText[$loop][0])
			{
				if ($KanjiCodeText[$loop][1] == 0)
				{
					# Single character match
					@_[0] .= $KanjiCodeText[$loop][2];
					$result = 1;
				}
				elsif ($second == $KanjiCodeText[$loop][1])
				{
					# Two character combo so skip past two characters
					@_[0] .= $KanjiCodeText[$loop][2];
					$result = 2;
				}
			}
			++$loop;
		}
	}

	return ($result);
}

#############################################################################
# state_name(<state to be converted>)
# returns an string for the full name of the state.
#
# Argument: State to display.
# Back: Full state name.
############################################################################
sub state_name {
	my $self = shift;		# Pointer to CGI structure
    my $state = shift;

    return $StateName{$state};
}

#############################################################################
# country_name(<country to be converted>)
# returns an string for the full name of the country.
#
# Argument: country to display.
# Back: Full state name.
############################################################################
sub country_name {
	my $self = shift;		# Pointer to CGI structure
    my $country = shift;

    return $CountryName{$country};
}

############################################################################
# bindump(<binary data to dump>)
# Will dump out the data as an HTML file
############################################################################
sub bindump {
	my $self = shift;

	my $binary = shift;
	my $max = $#{$binary} + 1;

	print $self->q->header,
		$self->q->start_html('Binary Dump of '.$max.' bytes'),
		$self->q->h1('Binary Dump of '.$max.' bytes');

	print "<TABLE BORDER=\"1\">\n";
	for (my $loop = 0; $loop < $max; $loop += 8)
	{
		print "<TR ALIGN=\"CENTER\" VALIGN=\"TOP\">";
		printf "<TD>%04X</TD>", $loop;
		printf "<TD>%02X</TD>", $binary->[$loop+0] & 0xFF;
		printf "<TD>%02X</TD>", $binary->[$loop+1] & 0xFF;
		printf "<TD>%02X</TD>", $binary->[$loop+2] & 0xFF;
		printf "<TD>%02X</TD>", $binary->[$loop+3] & 0xFF;
		printf "<TD>%02X</TD>", $binary->[$loop+4] & 0xFF;
		printf "<TD>%02X</TD>", $binary->[$loop+5] & 0xFF;
		printf "<TD>%02X</TD>", $binary->[$loop+6] & 0xFF;
		printf "<TD>%02X</TD>", $binary->[$loop+7] & 0xFF;
		print "</TR>\n";
	}
	print "</TABLE>\n";

	print $self->q->end_html;
}

#############################################################################
# vmfile_graffiti(<binary data>)
# returns an hash table derived from Graffiti File.
#
# Argument: Binary data of Graffiti File.
# Back: Hash Table containing Graffiti Elements.
############################################################################
sub	vmfile_graffiti{
	my $self = shift;		# Pointer to CGI structure

	my $binary = shift;

	my( $lCurPos ) = 0;
	my( $loop );
	my( $cbTitle, $cbComment, $cbMailAddress, $cbAlphaImage, $cbImage );
	my( %data );
		
	for ($loop = 0; $loop < 0x280; $loop++)	{
		$data{'header'} .= pack("c", $binary->[$lCurPos + $loop]);
	}
	$lCurPos += 0x280;
	
	$cbTitle = $binary->[$lCurPos] | ($binary->[$lCurPos + 1] << 8) | ($binary->[$lCurPos + 2] << 16) | ($binary->[$lCurPos + 3] << 24);
	$data{'titleSize'} = $cbTitle;
	$lCurPos += 4;
	if( $cbTitle ){
		for ($loop = 0; $loop < $cbTitle; $loop++)	{
			$data{'title'} .= pack("c", $binary->[$lCurPos + $loop]);
		}
		$lCurPos += $cbTitle;
	}

	$cbComment = $binary->[$lCurPos] | ($binary->[$lCurPos + 1] << 8) | ($binary->[$lCurPos + 2] << 16) | ($binary->[$lCurPos + 3] << 24);
	$data{'commentSize'} = $cbComment;
	$lCurPos += 4;
	if( $cbComment ){
		for ($loop = 0; $loop < $cbComment; $loop++)	{
			$data{'comment'} .= pack("c", $binary->[$lCurPos + $loop]);
		}
		$lCurPos += $cbComment;
	}

	$cbMailAddress = $binary->[$lCurPos] | ($binary->[$lCurPos + 1] << 8) | ($binary->[$lCurPos + 2] << 16) | ($binary->[$lCurPos + 3] << 24);
	$data{'mailSize'} = $cbMailAddress;
	$lCurPos += 4;

	if( $cbMailAddress ){
		for ($loop = 0; $loop < $cbMailAddress; $loop++)	{
			$data{'mail'} .= pack("c", $binary->[$lCurPos + $loop]);
		}
		$lCurPos += $cbMailAddress;
	}

	$cbAlphaImage = $binary->[$lCurPos] | ($binary->[$lCurPos + 1] << 8) | ($binary->[$lCurPos + 2] << 16) | ($binary->[$lCurPos + 3] << 24);
	$data{'alphaSize'} = $cbAlphaImage;
	$lCurPos += 4;
	if( $cbAlphaImage ){
		for ($loop = 0; $loop < $cbAlphaImage; $loop++)	{
			$data{'alpha'} .= pack("c", $binary->[$lCurPos + $loop]);
		}
		$lCurPos += $cbAlphaImage;
	}

	$cbImage = $binary->[$lCurPos] | ($binary->[$lCurPos + 1] << 8) | ($binary->[$lCurPos + 2] << 16) | ($binary->[$lCurPos + 3] << 24);
	$data{'imageSize'} = $cbImage;
	$lCurPos += 4;
	if( $cbImage ){
		for ($loop = 0; $loop < $cbImage; $loop++)	{
			$data{'image'} .= pack("c", $binary->[$lCurPos + $loop]);
		}
	}

	return	%data;
}

#############################################################################
# make_vmfile_graffiti(<Hash Table>)
# returns a new VMU file suitable for downloading.
#
# Argument: Hash Table data of Graffiti File.
# Back: binary string.
############################################################################
sub	make_vmfile_graffiti{
	my $self = shift;		# Pointer to CGI structure

	my $data  = shift;

	my( $binary, $number, $value, $clear, $left, $size );

	$binary = $data->{'header'};

	# Encode Title
	$number = $data->{'titleSize'};
	$number *= 2;
	$binary .= pack("C",($number & 0xFF));		# Convert number to 4 byte value
	$binary .= pack("C",(($number >> 8) & 0xFF));
	$binary .= pack("C",(($number >> 16) & 0xFF));
	$binary .= pack("C",(($number >> 24) & 0xFF));

	# Convert data into VMU compatible string
	foreach my $value(split(//,$data->{'title'}))
	{
		if ($value lt '0')
		{
			$binary .= pack("C",0x01);
			$binary .= ' ';
		}
		else
		{
			$binary .= pack("C",0x01);
			$binary .= $value;
		}
	}

	# Encode Comment
	$number = $data->{'commentSize'};
	$number *= 2;
	$binary .= pack("C",($number & 0xFF));		# Convert number to 4 byte value
	$binary .= pack("C",(($number >> 8) & 0xFF));
	$binary .= pack("C",(($number >> 16) & 0xFF));
	$binary .= pack("C",(($number >> 24) & 0xFF));

	# Convert data into VMU compatible string
	foreach my $value(split(//,$data->{'comment'}))
	{
		if ($value lt '0')
		{
			$binary .= pack("C",0x01);
			$binary .= ' ';
		}
		else
		{
			$binary .= pack("C",0x01);
			$binary .= $value;
		}
	}

	# Encode E-Mail Information
	$number = $data->{'mailSize'};
	$binary .= pack("C",($number & 0xFF));		# Convert number to 4 byte value
	$binary .= pack("C",(($number >> 8) & 0xFF));
	$binary .= pack("C",(($number >> 16) & 0xFF));
	$binary .= pack("C",(($number >> 24) & 0xFF));
	$binary .= $data->{'mail'};

	# Encode Alpha Channel Information
	$number = $data->{'alphaSize'};
	$binary .= pack("C",($number & 0xFF));		# Convert number to 4 byte value
	$binary .= pack("C",(($number >> 8) & 0xFF));
	$binary .= pack("C",(($number >> 16) & 0xFF));
	$binary .= pack("C",(($number >> 24) & 0xFF));
	$binary .= $data->{'alpha'};

	# Encode Image Information
	$number = $data->{'imageSize'};
	$binary .= pack("C",($number & 0xFF));		# Convert number to 4 byte value
	$binary .= pack("C",(($number >> 8) & 0xFF));
	$binary .= pack("C",(($number >> 16) & 0xFF));
	$binary .= pack("C",(($number >> 24) & 0xFF));
	$binary .= $data->{'image'};

	$size = ceil((length($binary) / 512.0)) * 512;
	$left = $size - length($binary);
	for ($clear = 0; $clear < $left; ++$clear)
	{
		$binary .= pack("C",0x00);
	}

	return	$binary;
}

#############################################################################
# decode_jgrData(<offset>, <binary data>)
# returns an binary array derived from JGR encoded the data stream.
#
# Argument: Binary data encoded by JGR.
# Back: Unencoded binary data (modifies passed variable).
############################################################################
sub decode_jgrData {
	my $self = shift;		# Pointer to CGI structure
 
	my $offset = shift;
    my $bindata = shift;

	my @dk = ();
	my ($i, $j, $shift, $nk, $size);

	# Defined Values
	my $SAVELOAD_KEY_NUM = 8;
	my $SYSTEM_SAVE_DATA = 1008;

	# Size of the actual ranking data
	$size = ($SYSTEM_SAVE_DATA - $SAVELOAD_KEY_NUM);

	# Read the last eight bytes
	for ($i = 0; $i < $SAVELOAD_KEY_NUM; $i++)
	{
		$dk[$i] = $bindata->[$offset + $size + $i];
		if ($dk[$i] & 0x80)
		{
			# Turn it into a signed number
			$dk[$i] = sprintf "%d", $dk[$i] |= 0xFFFFFF00;
			$dk[$i] = unpack('s', pack('S', $dk[$i]));
		}

		if ($dk[$i] == 0)
		{
			# An Error condition occurred.
			$bindata->[$offset + 0] = 0;
			$bindata->[$offset + 1] = 0;
			$bindata->[$offset + 2] = 0;
			$bindata->[$offset + 3] = 0;

			return;
		}
	}
	
	# Decode the binary data
	for ($i = $size - 1; $i >= 0; $i--)
	{
		$j = $i % $SAVELOAD_KEY_NUM;

		# 3bit Decoding
		$nk = ($dk[$j] >> 3) & 0x1f;
		$nk |= ($dk[$j] << 5) & 0xe0;
		if ($nk & 0x80)
		{
			# Turn it into a signed number
			$nk = sprintf "%d", $nk |= 0xFFFFFF00;
			$nk = unpack('s', pack('S', $nk));
		}


		$dk[$j] = $nk;
		$shift = $i + $dk[$j];
		#print $i.": ".$nk.", ".$shift.";";
		
		if ($shift < 0) {
			$shift += $size;
		}
		elsif ($shift >= $size) {
			$shift -= $size;
		}
		$nk = $bindata->[$offset + $i];
		$bindata->[$offset + $i] = $bindata->[$offset + $shift];
		$bindata->[$offset + $shift] = $nk ^ $bindata->[$offset + $i];
	}
	
    return;
}

############################################################################
# crc(<reference to array of bindata from data post>)
# returns a CRC value
############################################################################
sub crc {
	my $self = shift;		# Pointer to CGI structure

	my $bindata = shift;
	my $maxbyte = $#{$bindata} + 1;
    my $val=65535;
    my $CRCPOLY1=4129;      # 0x1021

    for(my $i=4;$i<$maxbyte;$i++) {
        $val ^= ($bindata->[$i]<<8);
        for(my $k=0;$k<8;$k++) {
            if(($val & 32768)!=0) {
                $val = ($val<<1)^$CRCPOLY1;
            } else {
                $val <<= 1;
            }
            $val &= 65535;
        }
    }
    $val ^= 65535;	# Bit Turnover
    $val &= 65535;

    return $val;
}

############################################################################
# CalculateChecksum(<position>, <size>, <reference to array of bindata from data post>)
# returns a CRC value
##################################################################################
sub CalculateChecksum()	{
	my $self = shift;		# Pointer to CGI structure

	my $pos = shift;		# Offset into binary data
	my $size = shift;		# Number of bytes to build the checksum on
	my $binary = shift;		# pointer to the checksum.

	my $sum = "00000000";
	my $key = 0;
	my $temp = 0;
	my $junk = 0;

	for($sum = 0; $size > 0; $size--)
	{
		$key = ($binary->[$pos] & 7);
		my $value = $binary->[$pos++] & 0xFF;
		if ($value & 0x80)
		{
			# Turn it into a signed number
			$value |= 0xFFFFFF00;
		}

		if ($key == 0)	{
			$temp = $self->bsub(hex($sum), $value);
			$sum = sprintf("%08X", $temp);
		}
		elsif ($key == 1)	{
			$temp = hex($sum) | $value;
			$sum = sprintf("%08X", $temp);
		}
		elsif ($key == 2)	{
			$temp = hex($sum) & $value;
			$sum = sprintf("%08X", $temp);
		}
		elsif ($key == 3)	{
			$temp = hex($sum) ^ $value;
			$sum = sprintf("%08X", $temp);
		}
		elsif ($key == 4)	{
			$temp = $self->bmul(hex($sum), $value);
			($junk,$temp) = $self->bdiv($temp, (hex("FFFFFFFF") + 1));
			$sum = sprintf("%08X", $temp);
		}
		elsif ($key == 5)	{
			$temp = $self->bdiv(hex($sum), $value);
			$sum = sprintf("%08X", $temp);
		}
		else {
			$temp = $self->badd(hex($sum), $value);
			($junk,$temp) = $self->bdiv($temp, (hex("FFFFFFFF") + 1));
			$sum = sprintf("%08X", $temp);
		}
#		printf "Value = \$%02X, key = %d, \$%s\n", $value, $key, $sum;
	}	# end of for

	return hex($sum);
}

#######################################################################
#CRC Calculation routine
#crcCalculateBuffer(<Number of bytes in the buffer, Current CRC value, The Buffer to calculate the CRC against>)
######################################################################
sub crcCalculateBuffer()
{
	my $self = shift;		# Pointer to CGI structure

	my $byteOffset = shift;
	my $nByteCount = shift;
	my $crcValue = shift;
	my $cBufferPtr = shift;
	
	my ($crc, $crcPolynomial, $crcData, $i, $j, @crcTable);

	# Do not change the Polynomial
	$crcPolynomial = 0xEDB88320;
	@crcTable = ();
	for ($i = 0; $i < 256; $i++) {
		$crc = $i;
		for ($j = 0; $j < 8; $j++) {
			if ($crc & 0x0001) {
				$crc = ($crc >> 1) ^ $crcPolynomial;
			}
			else {
				$crc = $crc >> 1;
			}
		}
		$crcTable[$i] = $crc;
	}

	#       Calculate the new CRC here :-
	$crc = $crcValue;
	for ($i = 0; $i < $nByteCount; $i++) {
		$crcData = ($crc ^ $cBufferPtr->[$byteOffset + $i]) & 0xFF;
		$crc = (($crc >> 8) & 0x00FFFFFF) ^ $crcTable[$crcData];
	}

	return $crc;
}

#####################################################################
# BinaryToFloat()
# Input : An array of bytes that represent a floating point number.
# Output : The decimal value which is equivalent of the input float.
#####################################################################
sub BinaryToFloat()
{
        my $self = shift;
        my $BinArr = shift;

        my($b0, $b1, $b2, $b3, $sign, $exponent, $non_exponent, $mult, $fraction, $pattern, $numerator, $bitset, @FractionBaseTable, $float, $x, $y, $z, $loop);

        my $str;

# Build the FractionBaseTable.
# It contains decreasing powers of 2 from 22 to 0.
        @FractionBaseTable = ();
        $FractionBaseTable[0] = 4194304;
        $FractionBaseTable[1] = 2097152;
        $FractionBaseTable[2] = 1048576;
        $FractionBaseTable[3] = 524288;
        $FractionBaseTable[4] = 262144;
        $FractionBaseTable[5] = 131072;
        $FractionBaseTable[6] = 65536;
        $FractionBaseTable[7] = 32768;
        $FractionBaseTable[8] = 16384;
        $FractionBaseTable[9] = 8192;
        $FractionBaseTable[10] = 4096;
        $FractionBaseTable[11] = 2048;
        $FractionBaseTable[12] = 1024;
        $FractionBaseTable[13] = 512;
        $FractionBaseTable[14] = 256;
        $FractionBaseTable[15] = 128;
        $FractionBaseTable[16] = 64;
        $FractionBaseTable[17] = 32;
        $FractionBaseTable[18] = 16;
        $FractionBaseTable[19] = 8;
        $FractionBaseTable[20] = 4;
        $FractionBaseTable[21] = 2;
        $FractionBaseTable[22] = 1;
        $FractionBaseTable[23] = 0;

#Read the 4 bytes
        $b0 = $BinArr->[0];
        $b1 = $BinArr->[1];
        $b2 = $BinArr->[2];
        $b3 = $BinArr->[3];

		# Check to see if everything is clear, then output zero
		if (($b0|$b1|$b2|$b3) == 0)	{
			return 0.0;
		}

# Take the sign bit : 7th bit of b3
        $sign = $b3 & 0x80;

# Calculate the exponent now.
        $x = ($b3 << 8) | $b2;
        $exponent = ($x >> 7) & 0xFF;
        $mult = $exponent - 127;

#Calculate the non-exponent component
        $y = ($b2 << 16) | ($b1 << 8) | $b0;
        $z = $y & 0x7FFFFF;
        $fraction = $z << 1;
        $pattern = $fraction;
        $numerator = 0;
        for ($loop = 0; $loop < 23; $loop++)    {
                $bitset = ($pattern & 0x800000);
                if ($bitset)    {
                        $numerator += $FractionBaseTable[$loop];
                }
                $pattern = $pattern << 1;
        }
        $non_exponent = ($numerator / 8388608) + 1.0;
        $float = ($non_exponent) * 2 ** $mult;
        if ($sign)      {
                $float *= (-1);
        }
        return $float;
}

############################################################################
# AUTOLOAD
# if the autoloaded function is a key in the %accessors
# global, we allow simple scalar get/set accessor functions.
#
# otherwise, we see if it's in another global hash or array - these
# are class data which AUTOLOAD provides read-only access to.
############################################################################
sub AUTOLOAD {
	my $self = shift;
	my $arg = shift;
	my $arg2 = shift;

	## need to fix this 'die' to something more polite sometime
	my $type = ref($self) || die "Unexpected error: $self is not an object";

	my $name = $AUTOLOAD;
	$name =~ s/.*://;

	if (exists $accessors{$name}) {
		if (defined $arg) {
			return $self->{$name} = $arg;
		} else {
			return $self->{$name};
		}
	} elsif (defined $arg) {
		no strict "refs";
		if (defined %{$name} && exists($name->{$arg})) {
			return $name->{$arg};
		} elsif ((defined @{$name}) && ($#$name >= $arg)) {
			return $name->[$arg];
		} 
	}
	return undef;
}
#------------------------ EOF --------------------------
1;

