# Heat.pm -- a general utility package for Heat CGI programs.
#
# $Id: Heat.pm,v 1.134 2000/05/07 23:42:57 kunerj Exp $
# Includes Channels changes (makes this incompatible with previous Heat.pm)

package Heat;

#use DBI;
use CGI;
use Exporter;
use Heat::Date;
use Heat::DbObject;
use URI::Escape;
#use Time::Timezone;
#use Time::Local;
#use Time::Wisely;
#use Sys::Hostname;

# Don't use hostnames here.  It breaks production servers when syslog
# attempts to do a reverse lookup on it.  Use only numeric IP addresses!
# This must preceed Sys:Syslog.
$Sys::Syslog::host = "127.0.0.1";
require Sys::Syslog;


#use strict;

use Heat::Size;
#use Heat::Player;
use Heat::SessionID;
use Heat::Env;

use vars qw( $Dbh @ISA $VERSION $DATABASE $USERNAME $PASSWORD
	     $TEMPLATE_ROOT $URL_ROOT $DRIVER @_WEEKDAYS @_MONTHS
	     @_SHORTDAYS @_SHORTMONS $HEAT_SERVER $OMI_SERVER
	     $PONG_SERVER $CHAT_SERVER
             $GAME $FORUM $DISCUSSION $CHAT_ROOM $ACTION
             $LADDER $affiliate_cookie_name 
	   );

@ISA = qw( Exporter );

$VERSION = 0.9;

# CONSTANTS... kinda.

*GAME = \1;
*FORUM = \3;
*DISCUSSION = \4;
*CHAT_ROOM = \5;
*ACTIVITY = \6;
*LADDER = \7;

# global variable $Dbh.

$Dbh = 0;

# set die() handler for Web mode


BEGIN {

    ## Catch compile-time errors as well
    $SIG{__DIE__} = \&DIE_handler;
   
    # set Environment    
    get Heat::Env;
}

$RUNTIME=1;
#sub DIE_handler ($;%) 
sub DIE_handler 
{
    my( $msg ) = shift @_;
    my( %data ) = @_;
	
	
	# Disconnect from DB, in case it has failed.
    Heat::disconnect();
    
    # Not in web mode?  Just die normal-like.
    if (!defined($ENV{GATEWAY_INTERFACE})) {
		print STDERR "OFF-LINE-MODE ERROR: $msg\n"
			if ($ENV{FASTCGI});
        #exit(55);
		return 56;
	}
	if (!defined $RUNTIME) {
	  print STDERR "COMPILE-TIME ERROR: $msg\n";
	  sleep(5);
	  return;
	}
    
    $data{state}->{return_to} = CGI::referer();
    $control->{no_cookie}=1;
    
    # If there was a "\n" at the end, show it to the user.
    # (unless it was a compile-time error)
    if ($msg !~ /line \d+/ && $msg !~ /compilation errors/) {
	  # User-readable error.  Don't log it 'cause it's not important.
	  $data{state}->{real_error} = $msg;
	  $data{state}->{error} = $msg;
	}
    else {
	my $hostname = "n/a";
	$hostname = `/usr/bsd/hostname` if (-x "/usr/bsd/hostname");
	$hostname = `/usr/bin/hostname` if (-x "/usr/bin/hostname");
	  $err_id = errlog( $msg );
	  errlog("HTTP_REFERER was $ENV{'HTTP_REFERER'} or maybe $ENV{'HTTP_REFERRER'}");
	  $data{state}->{real_error} = $msg;
	  $data{state}->{error} = "Whoa! Something is broken. " .
	    " Please wait a moment and retry your operation. " .
	    " If you continue to be plagued with troubles, please " .
		" report the following error to the HEAT support team: " .
		"'$err_id' ($hostname). Thanks. ";
    }
    
    # Unfortunately we have to hard code the path to the error content
    # because we reset $Heat::TEMPLATE_ROOT within the CGI scripts
    my ($ERROR_TEMPLATE);
    if (`hostname` =~ /mcp/ || `hostname` =~ /tick/) {
	# If we are running on the IP tool machine"
	$ERROR_TEMPLATE = "/usr/local/apache/htdocs/CustService/misc/error.html";
    } else {
	$ERROR_TEMPLATE = '/usr/local/www/heat/priv/content/misc/error.html';
    }
    
    # Run it through the error template
    if ( -e "$ERROR_TEMPLATE" ) { 
	if (`hostname` =~ /mcp/ || `hostname` =~ /tick/) {
	    $Heat::TEMPLATE_ROOT = '/usr/local/apache/htdocs/CustService/';
	} else {
	    $Heat::TEMPLATE_ROOT = '/usr/local/www/heat/priv/content/';
	}
      Heat::template( 'misc/error.html', control=> $control, %data ); 
    }
    else {
	Heat::errlog("Cannot find $ERROR_TEMPLATE");
	# Shit.
	print "Content-Type: text/plain\n";
	print "Status: 500 Server Error\n\n";
	print "Something really bad happened, and my error template" .
		" doesn't even exist. So, here:\n" . $data{state}->{error};
	print "\nand here:\n" . $data{state}->{real_error} . "\n";
    }
    
    
	# Normal CGI?  Exit with 77
    exit 77 if (!defined($ENV{FASTCGI}));

    # If we're in a fastcgi and the error isn't deadly, just "die-die"
    # This will break out of the main EVAL, and will wait for a new request.

    # However, if  the database connect fails then 
    # exit from the Fast CGI. We do this
    # to combat the following error:
    # ORA-12154: TNS could not resolve service name
    # We have found that restarting a fast CGI will resolve this problem.
    # jlowry (1/11/99)
    if ($msg =~ /^We are sorry/) {
      Heat::errlog("Exiting CGI because DB connect failed");
	exit 88;
    } elsif ($msg !~ /^EXIT/) {
	# To just die, we need only to return from here
	return; 	
    } else {
	# Acutally kill this process.  httpd will start a new one.
	# We don't want to swamp the webserver when things go bad.
	sleep 3;
	exit 88;
    }
}

sub errlog {

    my( $msg ) = @_;
    my( $err_id, $the_time, $line );

	my ($sec,$min,$hour,$mday,$mon,$year) = localtime(time);
    $the_time = sprintf "%.2d%.2d%.2d:%.2d:%.2d:%.2d",
		$year,$mon+1,$mday, $hour,$min,$sec;

    $err_id = "$the_time-$ENV{REMOTE_USER}";

	# Grab the last two elements of the script path
	my(@path) = split(/\/+/, $0);
	my($script) = pop @path;
	$script = (pop @path) . "/$script" if (@path);

	$line = "$script|$err_id|$msg";
	$line =~ s/\s*\n+\s*/  /g;	# remove newlines

    Sys::Syslog::openlog( "HEAT", 'pid', 'user' );
    Sys::Syslog::syslog( 'err', $line );
    Sys::Syslog::closelog();

    return $err_id;
}

sub connect(;$$$) {

    my( $database, $username, $password ) = @_;

    # defaults.

    $database = $DATABASE unless $database;
    $username = $USERNAME unless $username;
    $password = $PASSWORD unless $password;

	eval "use DBI;";
	die $@ if chomp $@;

	# Try to connect if we're not already connected.
    
if ( $Heat::Dbh ) {
       if ($Heat::Dbh->ping()){
           return $Heat::Dbh ;}
        Heat::disconnect();}

       $Heat::Dbh = DBI->connect( $database, $username, $password, $DRIVER );

	if (! $Heat::Dbh )
	{
		# Write to log, give user a message, and die.
		errlog("Can't connect to db $database with username $username and password $password and driver $DRIVER: $DBI::errstr");

		# If you change the error message below, then you
		# must also change the DIE handler which checks for this
		# message and exits if the database connection has failed
		my($msg) = "We are sorry for the inconvenience but HEAT is currently experiencing technical difficulties.  We apologize and we're working to remedy this situation as quickly as possible. For an update please go to: <a href=\"/pub/cgi-bin/consult/external_consult.cgi\">Online Consultation</a>. If Online Consultation isn't available, please try HEAT later.\n";

		# Die w/ this message if we're in CGI mode.
		die($msg) if defined $ENV{GATEWAY_INTERFACE};

		# Otherwise die with a shorter message.
		die("Can't connect to db $database: $DBI::errstr\n");
	}

    return $Heat::Dbh;
}

sub disconnect()
{
    my( $rc ) = 1;

    if ( $Heat::Dbh ) { $rc = $Heat::Dbh->disconnect; }

    undef $Heat::Dbh;
    return $rc;
}


sub wrapper_header { # mimics template

    $query = new CGI('');

    $header = $query->header( -expires => 'now',
                              -pragma => 'no-cache',
                              '-cache-control' => 'no-cache' );

    $header =~ s/^Content-type.*\n//mi;
    $header =~ s/^\s*\n//m;
    print $header;
}

###### HEAT::TEMPLATE(...) #######
# There is no problem so hard that it cannot be solved by a Perl
# script with sufficiently cryptic regexes.

sub template {

    my( $fname, %data ) = @_;
    local( *TEMPLATE );
    my( @lines );

    # Don't print header?  if they say so.
    my $no_header = $data{no_header};

    $data{state}->{server} = $HEAT_SERVER;
    $data{state}->{chat_server} = $CHAT_SERVER;
    $data{state}->{pong_server} = $PONG_SERVER;
    $data{state}->{flow_server} = $FLOW_SERVER;
    $data{state}->{sec_server} = $HEAT_SEC_SERVER;
    $data{state}->{session_id} = 
        ( $_session_id ) ? $_session_id : "NULL-SESSION-ID";
    $data{state}->{session_id_is_new} = 
        $Heat::_session_id_is_new && $_session_id;

    # Add Logical State Variable to test to see who is logged in or not
    # Added by C.L. Grassi  11/23/99.
    #-------------------------------------------------------------------

    $data{state}->{isLoggedin} = ($data{state}->{session_id} ne "NULL-SESSION-ID" || $data{state}->{session_id_is_new}) ? 1 : 0; 

    # End of change made by C.L. Grassi
    #----------------------------------

    # Allow them to specify relative- or absolute-paths.
    # Look in the TEMPLATE_ROOT first, then absolute (if starts with /)

    my $full_path = "$TEMPLATE_ROOT/$fname";
    $full_path = $fname if ($fname =~ /^\// && ! -e $full_path);

    open( TEMPLATE, "<$full_path" ) ||
      die( "Can't open template [$TEMPLATE_ROOT]/$fname: $!" );

    # We'll use this for include files later.
    my ($incl_path) = ($full_path =~ /^(.*)\/[^\/]+$/);  # default
    $incl_path = $incl_path || $data{INCLUDE_PATH};
    
    @lines = ( <TEMPLATE> );

    close( TEMPLATE );

    my( $query ) = new CGI('');

##Get state->affiliate.  If affiliate is passed by CGI parameter we set state->affiliate
##to the parameters and set the stage to pass the value as a cookie.  It should be noted
##that $affiliate_cookie_name is not only the name of the affiliate but also a boolean
##value to indicate whether the value must be set as cookie-- it only has a value when
##the affiliate changes, i.e. is passed as a CGI param.  If the affiliate is not sent as
##a CGI param $affiliate_cookie_name is *not* set but state->affiliate is assigned via the
##current cookie
my (@COOKIE_ARRAY);
if ($Heat::affiliate_cookie_name){
 $data{state}->{affiliate} = $affiliate_cookie_name;
 $affiliate_cookie = $query->cookie(-name => "affiliate",
                                         -value => $affiliate_cookie_name,
                                         -path => "/",
                                         -domain => "$Heat::HEAT_SERVER");

push (@COOKIE_ARRAY, $affiliate_cookie);
}
else{
 $data{state}->{affiliate} = $query->cookie("affiliate");
}
 
 

    # This templater might be used for email messages, so option for no header.
    # If the session ID is new then we want to print the cookie as part of
    # the HTTP header

    if (! $no_header && ($Heat::_session_id_is_new  || $Heat::session_from_param)) {

	my ($cookie) = $query->cookie( -name => "session_id",
				      -value => $_session_id,
				      -path => "/",
				      -domain => "$Heat::HEAT_SERVER");
push (@COOKIE_ARRAY, $cookie);
        if ( $data{control}->{dont_expire} ) {
	    print $query->header( -type => 'text/html',
			          -pragma => 'no-cache',
				 '-cache-control' => 'no-cache',
				 '-cookie' => \@COOKIE_ARRAY
			        );
        }
	elsif ($data{"control"}->{"cache"}) {
	    print $query->header( -type => 'text/html',
				 '-cookie' => \@COOKIE_ARRAY );
	}
        else {
	    print $query->header( -type => 'text/html',
			          -expires => 'now',
			          -pragma => 'no-cache',
			          '-cache-control' => 'no-cache',
				 '-cookie' => \@COOKIE_ARRAY
			        );
        }
     # We only need to set the affiliate cookie at the gateway
     # pages before the session ID is set
     } elsif ( ! $no_header && $Heat::affiliate_cookie_name) {


        if ( $data{control}->{dont_expire} ) {
	    print $query->header( -type => 'text/html',
			          -pragma => 'no-cache',
				 '-cache-control' => 'no-cache',
				 '-cookie' => \@COOKIE_ARRAY
			        );
        }
	elsif ($data{"control"}->{"cache"}) {
	    print $query->header( -type => 'text/html',
				 '-cookie' => \@COOKIE_ARRAY );
	}
        else {
	    print $query->header( -type => 'text/html',
			          -expires => 'now',
			          -pragma => 'no-cache',
			          '-cache-control' => 'no-cache',
				 '-cookie' => \@COOKIE_ARRAY
			        );
        }

    } elsif (! $no_header) {
        if ( $data{control}->{dont_expire} ) {
	    print $query->header( -type => 'text/html',
			          -pragma => 'no-cache',
				 '-cache-control' => 'no-cache' );
        }
	elsif ($data{"control"}->{"cache"}) {
	    print $query->header( -type => 'text/html' );
	}
        else {
	    print $query->header( -type => 'text/html',
			          -expires => 'now',
			          -pragma => 'no-cache',
			          '-cache-control' => 'no-cache' );
        }
    }

    _template_lines( \%data, \@lines, $incl_path );

}

# added 1/25/00 kunerj. Does the same as templater, but is for use
# in non-templated cgis that need the cookie, like player.cgi

sub get_affiliate_from_cookie {
  
  my( $query ) = new CGI('');
  my $affiliate;
  if ($Heat::affiliate_cookie_name){
    $affiliate = $affiliate_cookie_name;
  }
  else {
    $affiliate = $query->cookie("affiliate");
  }
  return $affiliate;
}

# Look up a file, read it in, and return a ref to array of lines
sub _read_include_file($$) 
{
    my ($path, $fname) = @_;


    # If $fname begins with $ it is interpreted as a Perl variable
    if ($fname =~ /^\s*\$/) {
	$fname = eval $fname;
    }

    # If the $path/$fname doesn't exist, and $fname starts with /, use abs path.

    my $full_path = "$path/$fname";
    $full_path = $fname if ($fname =~ /^\// && ! -e $full_path);

    my ($new_incl_path) = ($full_path =~ /^(.+)\/[^\/]+$/);

    # Open and read in the file.
    open( INCLUDE_F, "<$full_path")
        || die "Can't open template include file $full_path. $!";
    my @lines = <INCLUDE_F>;
    close INCLUDE_F;

    return (\@lines, $new_incl_path);
}
    
	

# The real template engine.  This is called recursively.
sub _template_lines {

    my( $data, $lines, $incl_path ) = @_;

    my( $BEGIN_MARKER ) =
     '<!-- BEGIN Autogen Bookmarking Code. Any changes between these lines will be lost -->';
    my( $END_MARKER ) =
     '<!-- END Autogen Bookmarking Code. Any changes between these lines will be lost -->';
    my( $BEGIN_APPLET ) =
     '<APPLET';
    my( $END_APPLET ) =
     '</APPLET>';
    my( $i, $repeat_cnt );
    my( @repeat_lines, $if_cnt, @if_lines );
    my( $thing, $list, $short_name, $not, $name );
    my( $preview ) = $data->{control}->{preview_mode};

    for ( $i = 0; $i <= $#$lines; $i++ ) {

	$_ = $lines->[$i];

	next if ( $preview && ( /$BEGIN_MARKER/ .. /$END_MARKER/ ) );
	next if ( $preview && ( /$BEGIN_APPLET/ .. /$END_APPLET/ ) );

	# Parse the new %%token%%
    s/\%\%([\w\.\[\]\d]+)\%\%/_resolve($data,$1)/eg;


	# Include a file?
	if ( ($name) = /^\s*\#\#INCLUDE:(.*)\#\#/ ) {
	    my ($incl_lines, $new_incl_path) = 
	        _read_include_file($incl_path, $name);

	    _template_lines( $data, $incl_lines, $new_incl_path );

	}


	# Repeat over a list?
	elsif ( ( $name ) = /^\s*\#\#REPEAT:([\w\.\[\]\d]+)\#\#/ ) {
	    
	    $list = _resolve( $data, $name );
	    $short_name = _get_short_name( $name );
	    
	    # if you use empty lists, you suck.  Just so you
	    # know.
	    
	    if ( not defined $list ) { $list = [] }
	    
	    $repeat_cnt = 1;
	    
	    $#repeat_lines = -1;
	    
	  REPEAT_LINE: for ( $i++; $i <= $#$lines; $i++ ) {
		
		$_ = $lines->[$i];
		
		if (/^\s*\#\#REPEAT:.*?\#\#/ ) {
		    $repeat_cnt++;
		}
		elsif (m+^\s*\#\#/REPEAT\#\#+ ) {
		    $repeat_cnt--;
		    last REPEAT_LINE if $repeat_cnt == 0;
		}
		
		push @repeat_lines, $_;
	    }
	    
	    my $save_this = $data->{$short_name};  # avoid name collision.
	    foreach $thing ( @{$list} ) {
		
		$data->{$short_name} = $thing;
		
		_template_lines( $data, \@repeat_lines, $incl_path );
	    }
	    
	    delete $data->{$short_name};

	    $data->{$short_name} = $save_this  if defined $save_this;

	}

	# Do an IF clause?
	elsif ( ( $name ) = /^\s*\#\#IF:([\w\.\[\]\d]+)\#\#/ ) {
	    
	    ( $not, $name ) = _get_not( $name );
	    
	    $obj = _resolve( $data, $name );
	    
	    $if_cnt = 1;
	    
	    $#if_lines = -1;
	    
	  IF_LINE: for( $i++; $i <= $#$lines; $i++ ) {
		
		$_ = $lines->[$i];
		
		if ( /^\s*\#\#IF.*?\#\#/ ) {
		    $if_cnt++;
		}
		elsif ( m+^\s*\#\#/IF\#\#+ ) {
		    $if_cnt--;
		    last IF_LINE if $if_cnt == 0;
		}
		push @if_lines, $_;
	    }
	    
	    if ( ( $not ) ? !$obj : $obj  ) {
		_template_lines( $data, \@if_lines, $incl_path );
	    }
	}

	# Else just replace proper tokens.
	else {
	    s/\#\#([\w\.\[\]\d]+)\#\#/_resolve($data,$1)/eg;

	    if ( $preview ) {
		# we have to hack up the image sources.
		s@\.\./\.\./images@$IMAGE_URL_ROOT@g;
		# and the links.
		#s@href\s*=\s*[\'\"]?.*?[\'\"]?@href=\"javascript:alert('links not active in preview mode')\"@gi;
		# and secure connections.
		s/https/http/;
		# and public graphics
		s@pub/graphics@CustService/Images/pub-graphics@g;

	    }

	    s/&hsh;/\#/g;
	    # added to make all links viewable in heat client
	    # external with hax0r mpath httpx protocol. :P
	    if($data->{heat_client}->{display_httpx} == '1') {
		s/http\:/httpx\:/g;
	    }
	    print;
	}
    }
}


sub _resolve {
    
    my( $data, $name ) = @_;
    my( @fields ) = split( /\./, $name );

    my( $field );
    my( $last_field ) = "<top>";
    my( $obj ) = $data;

    while ( $field = shift @fields ) {

	if ( not defined $obj ) { $obj = {} }
	if ( not ref $obj ) { die "'$last_field' is not an Object!" }

	$last_field = $field;

	if ( $field =~ s/\[(\d+)\]// ) {
	    $obj = $obj->{$field}[$1];
	}
	else {
	    $obj = $obj->{$field};
	}

    }

    return $obj;
}

sub _get_short_name {

    my( $name ) = @_;
    my( @fields ) = split( /\./, $name );
    my( $short_name ) = pop @fields;

    return $short_name;
}

sub _get_not {

    my( $name ) = @_;
    my( @fields ) = split( /\./, $name );
    my( $not, $real_name );

    $not = ( $fields[0] eq "NOT" ) ? 1 : 0;

    if ( $not ) { shift @fields }

    $name = join( ".", @fields );

    return ( $not, $name );
}

sub die_on_err {

    $_db_die = $Heat::Dbh->{RaiseError};
    
    $Heat::Dbh->{RaiseError} = 1;
}

sub restore_err {
    
    $Heat::Dbh->{RaiseError} = $_db_die;
}

sub text_to_html($) {


    my( $text ) = shift @_;
    
    $text =~ s/&/&amp;/g;
    $text =~ s/</&lt;/g;
    $text =~ s/>/&gt;/g;
    $text =~ s/\"/&quot;/g;
    $text =~ s/\#/&hsh;/g;
    $text =~ s/\'/&\#039;/g;
    $text =~ s/\n\n/<P>/g;
    $text =~ s/\n/<BR>/g;

    return $text;
}

sub html_to_text($) {

    my( $text ) = shift @_;
    
    $text =~ s/<P>/\n\n/g;
    $text =~ s/<BR>/\n/g;

    $text =~ s/&amp;/&/g;
    $text =~ s/&lt;/</g;
    $text =~ s/&gt;/>/g;
    $text =~ s/&quot;/\"/g;
    $text =~ s/&\#039;/\'/g;
    $text =~ s/&hsh;/\#/g;
    
    return $text;
}

## skip session_id validation, but pass along session_id
sub skip_session_id_check($) {

    my( $query ) = @_;

    undef $_username;
    undef $ENV{REMOTE_USER};

    $_session_id = $query->param( "session_id" );
    $Heat::_session_id_is_new = 0;

    return undef;
}

## This validates a session_id
sub check_referrer($;$) {

    my( $query , $_dont_die) = @_;

    # Use cookies if no session ID supplied to authenticate
    # (or maybe that should be the other way round)
    my( $session_id ) = $query->param( "session_id" );
	#If session_id came from parameter we need to note that:
	if ($query->param( "session_id" )){
		$Heat::session_from_param = 1;
	}else{undef $Heat::session_from_param;}

    if (! $session_id || $session_id eq "NULL-SESSION-ID") {
	$session_id = $query->cookie("session_id");
    }

# was there "affiliate=xxx" if so we must record it for the template sub

   if ($query->param( "affiliate" )){
     $Heat::affiliate_cookie_name = lc $query->param( "affiliate");
  }
	else{
	 undef $Heat::affiliate_cookie_name;
  }
    
	local $dont_die = $_dont_die;

    undef $_session_id;
    undef $_username;
    undef $ENV{REMOTE_USER};
    undef $Heat::DEGREES;
	$Heat::_session_id_is_new = 0;

	# If we reject them, refer them back to login.cgi, and log this occurance.
	sub reject_them($) {

		return $_[0] if $dont_die;

		my %state;
		$state{error_txt} = 
		  CGI::escape("You are not currently logged in to Heat. " .
			      "Please enter your name and password.");

		# If the user is trying to play a game without logging
		# in then they must have come from a channels page and
		# hence we can redirect to the special channels login page
		my(@path) = split(/\/+/, $0);
		my($script) = pop @path;
		$script = (pop @path) . "/$script" if (@path);
		my $esc_script = pop (@path) . "/$script";
		$esc_script = uri_escape($esc_script . "?" . $ENV{"QUERY_STRING"},"=?&");
		my $dest; # the distination will be added to index.cgi parameter so the inter
			  # mediate login page can be more generic ngb063099 
		if ( $esc_script =~ /home/) {
			$dest="home";
		} elsif ($esc_script =~ /chat/) {
			$dest="chat";
		} elsif ($esc_script =~ /commerce/) {
			$dest="store";
		}

		if ($dest) {
		    if ($esc_script =~ /affiliate%3D([^&]+)/) {
	        	Heat::redirect("/pub/cgi-bin/login/index.cgi?script=$esc_script&channel=1&dest=$dest&affiliate=$1");
		      } else {
			Heat::redirect("/pub/cgi-bin/login/index.cgi?script=$esc_script&channel=1&dest=$dest");
		      }
		} else {
		    if ($esc_script =~ /affiliate%3D([^&]+)/) {
	        	Heat::redirect("/pub/cgi-bin/login/index.cgi?script=$esc_script&channel=1&affiliate=$1");
		    } else {
	        	Heat::redirect("/pub/cgi-bin/login/index.cgi?script=$esc_script&channel=1");
		      }
		}
		errlog("Rejected session_id from $ENV{REMOTE_ADDR}: $_[0]");

    	undef $_session_id;
    	undef $_username;
    	undef $ENV{REMOTE_USER};

		local $SIG{__DIE__} = undef;
    	die "Rejected session_id: $_[0]\n";
	}

    reject_them "No session_id found."
    	if !defined $session_id;

    reject_them "Zero length."
    	if length($session_id) == 0;

    my( $sesh ) = new Heat::SessionID;

    reject_them "Decode failed."
       if ! $sesh->decode( $session_id );


    # Save the username, so we get more better logging.
    $_username = $sesh->{username};
    $ENV{REMOTE_USER} = $sesh->{username};

    reject_them "Timestamp is zero."
		if (int $sesh->{timestamp} == 0);

    reject_them "Username field is empty."
		if ($sesh->{username} !~ /./);

    my( $now ) = time;

	# Expired?  If they're connected, give them a new timestamp.
    if (( $now - $sesh->{timestamp} > $EXPIRE_LIMIT ) && (! $dont_die)) {
		Heat::connect;

		eval "use Heat::User;";
		die $@ if chomp $@;

		my $user = get Heat::User( $sesh->{username} );

		reject_them "No such user '$sesh->{username}'"
			if !$user;

		my $dt = $now - $sesh->{timestamp};
		reject_them "Not online after timeout ($dt seconds)."
			if  !$user->{online};

		# New session_id
		$sesh->{timestamp} = $now;
		$sesh->{degrees} = $Heat::DEGREES = $user->{degrees};
		$session_id = $sesh->encode();
		#errlog("New session_id generated after online-timout.");

		# Set this so the templater can pass it on to the javascript
		$Heat::_session_id_is_new = 1;
    }

    # OK, things are cool.

    $_session_id = $session_id;

    return undef;
}

sub redirect($;$) {

    my( $location, $secure ) = @_;

    print "Location: http://$HEAT_SERVER/$URL_ROOT$location\n\n";
}

sub secure_redirect($;$) {

    my( $location, $secure ) = @_;

    print "Location: https://$HEAT_SEC_SERVER/$URL_ROOT$location\n\n";
}

# cleggp980903:  redirect_external takes a full URL and simply sends out
# a full redirect.
sub redirect_external ($;) {
  my ($location) = @_;
  print "Location:  $location\n\n";
}

sub no_response {

    print "Status: 204 No Content\n\n";

}

sub log($$) {
    
    my( $res_id, $user_id ) = @_;
    my( $sth );

    $sth =
      $Heat::Dbh->prepare( "BEGIN Heat_Heat.Log_Res_Use( $res_id, $user_id ); END;" ) ||
	die Heat::errstr();
    
    $sth->execute || die Heat::errstr();
    $sth->finish || die Heat::errstr();
    
    
    return 1;
}

sub err() { $DBI::err }
# Remove trailing newline from error (so that DIE does the right thing)
sub errstr() { 
    my $e = $DBI::errstr;
    $e =~ s/\n+$//;   
    return $e;
}

## Commit any pending writes to the database.
## Call this after "execute" of any write to DB.
sub db_commit()
{
	$Heat::Dbh->commit || die "Can't commit: ". Heat::errstr();
}



sub return_to(;$) {

    my( $default ) = @_;

    my( $referrer ) = CGI::referer();

    return $referrer || $default || "/cgi-bin/misc/home.cgi";
}

# not for public consumption.

sub _is_online {

    my( $player );

    Heat::connect;

	eval "use Heat::Player;";
	die $@ if chomp $@;
    
    $player = get Heat::Player( $ENV{REMOTE_USER} );
    
    return $player->{online};
}

sub insert_realm_DSAKey_certificate($$){
    my ($private_key, $certificate)=@_;
    my $_NUM_DAYS =365;
    my $_TYPE_OF_DURATION = 'D'; #D for day.

my $INSERT_REALM_KEY_CERTIFICATE = "BEGIN HEAT_DSA.Insert_realm_key(:1, :2, :3, :4); END;";

my $ch = $Heat::Dbh->prepare($INSERT_REALM_KEY_CERTIFICATE)||die Heat::errstr();

    $ch->bind_param(1, $private_key);
    $ch->bind_param(2, $certificate);
    $ch->bind_param(3, $_NUM_DAYS);
    $ch->bind_param(4, $_TYPE_OF_DURATION);
    $ch->execute() || die Heat::errstr();
    $ch->finish() || die Heat::errstr();

Heat::db_commit();

}
sub delete_realm_key_certificate{
    my $_DAYS_EXPIRATION = 7;

my $DELETE_EXPIRED_KEY = "BEGIN HEAT_DSA.Delete_expired_realm_keys(:1); END;";
my $ch = $Heat::Dbh->prepare($DELETE_EXPIRED_KEY)||die Heat::errstr();
    $ch->bind_param(1, $_DAYS_EXPIRATION);
    $ch->execute() || die Heat::errstr();
    $ch->finish() || die Heat::errstr();

Heat::db_commit();
}

sub get_realm_DSAKey_certificate {
    my ($private_key, $certificate, $key_expiration);
    my $realm={};
    my  @list=[];
    my $SQL = <<"END_OF_SELECTSQL";
SELECT private_key , certificate, expiration 
FROM realm_dsa_key 
WHERE (sysdate - expiration) < 7 
ORDER BY  expiration DESC 
END_OF_SELECTSQL


my( $sth ) = $Heat::Dbh->prepare( $SQL )||die Heat::errstr();
$sth->execute() || die Heat::errstr();
 
while (($private_key,$certificate, $key_expiration) = $sth->fetchrow )
{
    $realm->{private_key}=$private_key;
    $realm->{certificate}=$certificate;
    $realm->{key_expiration}=$key_expiration;
    
push @list, $realm;
$realm={};
}

$sth->finish();

return @list;
}


sub get_permissions {

    my( $object_id, $object_type_id ) = @_;
    my( $list ) = [];
    my( $name, $id, $code, $sign, $desc );

    # Get group permissions

    my( $sth ) = $Heat::Dbh->prepare( <<"END_OF_SQL" ) || die Heat::errstr();
SELECT g.group_name, g.group_id, p.permission_code, p.permission_sign,
     pc.description
FROM usergroup g, permission p, permission_code pc
WHERE p.object_type_id = :object_type_id
  AND p.object_id = :object_id
  AND g.group_id = p.group_id
  AND p.permission_code = pc.permission_code
END_OF_SQL

    $sth->bind_param( ":object_id", $object_id );
    $sth->bind_param( ":object_type_id", $object_type_id );

    $sth->execute();

    while ( ( $name, $id, $code, $sign, $desc ) = $sth->fetchrow ) {

	if ( $sign eq '-' ) {
	    $desc = "Can\'t " + $desc;
	}

	$perm = new Heat::Permission( object_id => $object_id,
				      object_type_id => $object_type_id,
				      name => $name,
				      id => $id,
				      code => $code,
				      sign => $sign,
				      desc => $desc,
				      ptype => 'group'
				    );

	push @$list, $perm;
    }

    $sth->finish;

    # Get user permissions

    $sth = $Heat::Dbh->prepare( <<"END_OF_SQL" ) || die Heat::errstr();
SELECT pl.username, pl.player_id, p.permission_code, p.permission_sign,
    pc.description
FROM valid_player pl, permission p, permission_code pc
WHERE p.object_type_id = :object_type_id
  AND p.object_id = :object_id
  AND pl.player_id = p.player_id
  AND p.permission_code = pc.permission_code
END_OF_SQL

    $sth->bind_param( ":object_id", $object_id );
    $sth->bind_param( ":object_type_id", $object_type_id );

    $sth->execute();

    while ( ( $name, $id, $code, $sign, $desc ) = $sth->fetchrow ) {

	if ( $sign eq '-' ) {
	    $desc = "Can\'t " + $desc;
	}

	$perm = new Heat::Permission( object_id => $object_id,
				      object_type_id => $object_type_id,
				      name => $name,
				      id => $id,
				      code => $code,
				      sign => $sign,
				      desc => $desc,
				      ptype => 'player'
				    );

	push @$list, $perm;
    }
	$sth->finish;

    return $list;
}

# make sure we disconnect at the end.

END {
    my $w = $^W; $^W=0; # turn off warnings for a sec.
    undef $SIG{__DIE__}; # global cleanup causes some eval's, bummer.
    $^W = $w;
    Heat::disconnect;
}


package Heat::Permission;

@ISA = qw( Heat::DbObject );

1; # the loneliest number.

__END__

=head1 NAME

Heat - general utility package for other Heat::* packages to use.

=head1 SYNOPSIS

    use Heat;
 
    Heat::check_referrer; # will do magic redirection if not OK.
 
    Heat::connect;
 
    # get resource ID of this resource.
 
    $res = Heat::res_id( $ENV{SCRIPT_NAME} );
 
    # log this use.
 
    Heat::log( $res, $user->{id} );
     
    # note: don't do any DBI stuff in your CGI!
 
    $mystmt = $Heat::Dbh->prepare( 'execute whatever( :1, :2, :3 )' );
 
    Heat::template( 'myfile.html', user => $user );
 
    Heat::disconnect;
 
=head1 DESCRIPTION

The Heat package provides some common stuff for the other Heat::*
packages to use.  Mainly, this involves making one database connection
for all to use.

Note that this package is procedural rather than OOPy.

=head1 CONTENTS

=over 4

=item VARIABLES

=over4

=item $Dbh

This is a global database handle.  It's initialized when the Heat module is
loaded, so it should be always available.  CGI programs should leave it alone;
modules should use it exclusively.

=back

=item SUBS

=over 4

=item connect( [$database [,$username [,$password] ] ] )

Connects to the database $database, using $username and $password.
Passes these arguments through to DBI->connect(), so other
arguments that fit in these slots are acceptable.

All arguments are optional.  If an argument is not provided,
the default value is used.

Returns undef if the connection is not made.

=item disconnect()

Disconnects the global database handle, and undef's the variable.
Don't use $Dbh after disconnect'ing, or you will be damaged.

=item template( $filename, %data )

Squishes the data in hash %data into HTML template file $filename.
Note that since Perl treats sub arguments like a list, and hashes can
be initialized from a list, you can just pass in your arguments by
name, like this:

  Heat::template( '../home/personalpage.html', 
            user => $user, mailbox => $mailbox );

Or, you can do it another way.  Just don't waste a lot of time.

=item redirect( $url )

Does a CGI redirection of the user to another page, referred to by
$url.

=item check_referrer( [$referrer] , $dont_die?)

Checks to see if the URL passed is OK.  If not, then this sub will
output sufficient Location CGI headers to send the user back up to the
Page-Right-After-Login Page.  It will then call die (a local, simple
version of die).  If you set $dont_die to true (1), check_referrer
will return an error message and not die or redirect, or else
return undef if there is no error (session_id is good).  This option
is useful if you want to catch session_id errors and not actually
redirect the user.

If $referrer isn't provided, $ENV{HTTP_REFERER} will be used.

Note that the HTTP protocol specification misspells "referrer":

http://www.m-w.com/cgi-bin/mweb?va=referrer

as "referer":

http://www.ics.uci.edu/pub/ietf/http/rfc1945.html#Referer

Bothersome, but we use the correct spelling in this function.

=item res_id( [$res_url] )

Returns the resource id of the resource at $res_url to the caller.

If no $res_url is provided, returns the resource id of $ENV{SCRIPT_NAME}.

=item log( $res_id, $user_id )

Logs that user $user_id has accessed resource $res_id.  The resource can
be a physical resource, like a script, or a 'virtual' resource, like a forum,
chat room, game or player.

=item get_permissions( $object_id, $object_type_id )

Retrieves a reference to a list of L<Heat::Permission> objects that
represent all permissions on the object of type $object_type_id with
id $object_id.

=back

=back

=head1 CLASSES

=head2 Heat::Date

This is a date.  You can construct one using the magical constructor
that converts run-of-the-mill everyday dates into cool timezone-aware
dates.  Neato!

=over 4

=item ATTRIBUTES

=over 4

=item real_time

The real, honest-to-god time it is.

=item tz

A displayable value of the time zone, like "PST".

=item user_time

The time it would really have to be in Greenwich to get the same hour,
minute, dot dot dot as you would B<now> in tz.  Really, just the
real_time plus the timezone offset of tz.  Clear as mud, right?

=item sec

Seconds of this date, with leading zero.

=item min

Minutes, with leading zero.

=item hour24

Hour in 24-hour format, like sicko pervert French communists use.

=item hour

Hour in 12-hour format, the way all decent Americans do it.

=item ampm

The string 'AM' or 'PM', depending on if it's AM or PM
