#!/usr/bin/perl -I.
# Dreamcast Version
# Rush 2049
# World Rankings Displayer


use strict;
use integer;
use CGI;
use POSIX	'strftime';

	
	require "bigint.pl";

	# Variables
# -----------------------------------------------------
	use vars qw(@base64 @dp_base64 $Stage $Mode @StageName @GameType @LevelName @StageTbl @CharactersTbl @RankTbl @BINDATA $BINARY);
	
	use vars qw($errorURL);


# encryption table
@base64    = ('A','B','C','D','E','F','G','H','I','J',
              'K','L','M','N','O','P','Q','R','S','T',
              'U','V','W','X','Y','Z','a','b','c','d',
              'e','f','g','h','i','j','k','l','m','n',
              'o','p','q','r','s','t','u','v','w','x',
              'y','z','0','1','2','3','4','5','6','7',
              '8','9','+','/','=');

@dp_base64 = ('A','Z','O','L','Y','N','d','n','E','T',
              'm','P','6','c','i','3','S','z','e','9',
              'I','y','X','B','h','D','g','f','Q','q',
              '7','l','5','b','a','t','M','4','r','p',
              'K','J','j','8','C','u','s','x','R','F',
              '+','k','2','V','0','w','U','G','o','1',
              'v','W','H','/','=');

@StageName = ( 'Shibuya-Cho',
               'Kogane-Cho',
               'Benten-Cho',
               'Bantam Street',
               'Grind Square',
             );

@GameType = (  'Story',
			   'Jet Crush',
               'Jet Technique',
               'Jet Graffiti',
             );

@LevelName = ('Shibuya GG','Love Trap','Love Attack','Explosion','',
			  '1st','2nd','3rd','4th','5th',
			  '1st','2nd','3rd','4th','5th',
			  '1st','2nd','3rd','4th','5th',

			  'The Monster of Kogane','No. 540','Kogane Circus','Fight or Flight','',
			  '1st','2nd','3rd','4th','5th',
			  '1st','2nd','3rd','4th','5th',
			  '1st','2nd','3rd','4th','5th',

			  'Benten Boogie','Graffiti High','Noise Reduction','Benten Burning','Behind the Mask',
			  '1st','2nd','3rd','4th','5th',
			  '1st','2nd','3rd','4th','5th',
			  '1st','2nd','3rd','4th','5th',

			  'Tag or Die','','','','',
			  '1st','2nd','3rd','4th','5th',
			  '1st','2nd','3rd','4th','5th',
			  '1st','2nd','3rd','4th','5th',

			  'Rock\' N Roll Grind','','','','',
			  '1st','2nd','3rd','4th','5th',
			  '1st','2nd','3rd','4th','5th',
			  '1st','2nd','3rd','4th','5th',
			 );

# Stage Related Data
@StageTbl = (
		# key => [ Stage Name, Stage Graphic, VMU File ]
		[ "/images/maptext_shibuya.gif", "/images/maps_shibuya.gif", "JETGRIND_SYS"],
		[ "/images/maptext_kogane.gif", "/images/maps_kogane.gif", "JETGRIND_SYS"],
		[ "/images/maptext_benten.gif", "/images/maps_benten.gif", "JETGRIND_SYS"],
		[ "/images/maptext_bantam.gif", "/images/maps_bantam.gif", "JETGRIND_SYS"],
		[ "/images/maptext_grindsq.gif", "/images/maps_grindsq.gif", "JETGRIND_SYS"],
        );

@CharactersTbl = (
				# key => [ Character Name, Character Graphic, Hex Value ]
				[ "Beat",			"/images/char_beat.gif",	"0" ],
				[ "Gum",			"/images/char_gum.gif",		"1" ],
				[ "Tab",			"/images/char_tab.gif",		"2" ],
				[ "Combo",			"/images/char_combo.gif",	"3" ],
				[ "Slate",			"/images/char_slate.gif",	"4" ],
				[ "Piranha",		"/images/char_piranha.gif",	"5" ],
				[ "Cube",			"/images/char_cube.gif",	"6" ],
				[ "Mew",			"/images/char_mew.gif",		"7" ],
				[ "Yo-Yo",			"/images/char_yoyo.gif",	"8" ],
				[ "Garam",			"/images/char_garam.gif",	"9" ],
				[ "Poison Jam",		"/images/char_poison.gif",	"A" ],
				[ "Love Shocker",	"/images/char_love.gif",	"B" ],
				[ "Noise Tanks",	"/images/char_noise.gif",	"C" ],
				[ "Goji",			"/images/char_goji.gif",	"D" ],
				[ "Pots",			"/images/char_pots.gif",	"E" ],
				);

@RankTbl = (
			# key => [ Class Name, Class Graphic ]
			[ "Pedal!",		"/images/rank_pedal.gif" ],
			[ "Motor!",		"/images/rank_motor.gif" ],
			[ "Engine!",	"/images/rank_engine.gif" ],
			[ "Turbo!",		"/images/rank_turbo.gif" ],
			[ "Nitro!",		"/images/rank_nitro.gif" ],
			[ "Jet!",		"/images/rank_jet.gif" ],
		 );

######################################################################

	$errorURL = "rank.html";

	@BINDATA = &read_vmu_upload();

	# Retrieve the cookie information
	my %cookies = &getCookies; 

	my $userid = $cookies{'login_userID'};

	if ($userid eq "")
	{
		&error("[upload_rank] Please login prior to uploading your World Ranking File.  Use the login button on the Home Page.", "home.html");
	}

	# Check to see if file was uploaded
	if ($#BINDATA > 0)
	{
		##############################################################
		# Data structure decode begin
		my ($ptr, $offset);
		my ($guid, $volume, $gameclear);
		my (@rank_point, @rank_rank, @rank_player);
		
		&decode_jgrData(0x280, \@BINDATA);
		
		$offset = 0x280;
		
		
		$ptr = 0;
		$guid = $BINDATA[$offset + $ptr] | ($BINDATA[$offset + $ptr + 1] << 8) | ($BINDATA[$offset + $ptr + 2] << 16) | ($BINDATA[$offset + $ptr + 3] << 24);
		$ptr += 4;

		###########################
		# if($guid != 26060415) {			
			# &error("[upload_rank] Checksum Error!<BR>World Ranking data is corrupted.", "rank.html");
		# }
		###########################

		my @tempArr1 = ($BINDATA[$offset + $ptr], $BINDATA[$offset + $ptr + 1], $BINDATA[$offset + $ptr + 2], $BINDATA[$offset + $ptr + 3]);
		$volume = &BinaryToFloat(\@tempArr1);
		$ptr += 4;

		$gameclear = $BINDATA[$offset + $ptr] | ($BINDATA[$offset + $ptr + 1] << 8) | ($BINDATA[$offset + $ptr + 2] << 16) | ($BINDATA[$offset + $ptr + 3] << 24);
		$ptr += 4;
	
		for (my $i = 0; $i < (5 * 4); $i++)
		{
			for (my $j = 0; $j < 5; $j++)
			{
				$rank_point[$i][$j] = $BINDATA[$offset + $ptr] | ($BINDATA[$offset + $ptr + 1] << 8) | ($BINDATA[$offset + $ptr + 2] << 16) | ($BINDATA[$offset + $ptr + 3] << 24);
				$ptr += 4;
				if (($i == 1) || ($i == 5) || ($i == 9) || ($i == 13) || ($i == 17))
				{
					$rank_point[$i][$j] = $rank_point[$i][$j] / 100.0;
				}
			
				$rank_rank[$i][$j] = $BINDATA[$offset + $ptr];
				$ptr += 1;
			
				$rank_player[$i][$j] = $BINDATA[$offset + $ptr];
				$ptr += 1;
			
				# Possible Padding
				$ptr += 1;
			
				# Possible Padding
				$ptr += 1;
			}
		}
		
		# Set up variables needed for output page
		#########################################################
		my ($temp, $sName, $lName, $gName, $offset, $text, $value, $char, $rank);

		if(!defined($StageName[$Stage])) {
			&error("[upload_rank] Invalid stage was chosen.<BR>Please select a different stage.", $errorURL);
		}

		# All Modes use only entry "1st" except Story Mode.

		# Story Mode
		if (($Mode == 0) || ($Mode == 10) || ($Mode == 20) || ($Mode == 30) || ($Mode == 40)) {
			# This mode is not used
			my $val = (($Stage * (4 * 5)) + ($Mode / 10));
			if($LevelName[$val] ne '') {
				$sName = $StageName[$Stage];
				$lName = $LevelName[$val];
				$gName = $GameType[0];
				$text = "Points";
				$offset = ($Mode / 10);
				$value = $rank_point[(($Stage * 4) + 0)][$offset];
				$rank = $rank_rank[(($Stage * 4) + 0)][$offset];
				$char = $rank_player[(($Stage * 4) + 0)][$offset];
			} else {
				&error("There is no Story Mode in $StageName[$Stage] Level ".(($Mode/10) + 1).".", $errorURL);
			}
		} elsif($Mode == 1) {			# Jet Crush Mode
			my $val = (($Stage * (4 * 5)) + ($Mode * 5));
			if($LevelName[$val] eq '1st') {
				$sName = $StageName[$Stage];
				$lName = "";
				$gName = $GameType[$Mode];
				$text = "Time";
				$offset = 0;
				$value = &CalcTime($rank_point[(($Stage * 4) + 1)][$offset]);
				$rank = $rank_rank[(($Stage * 4) + 1)][$offset];
				$char = $rank_player[(($Stage * 4) + 1)][$offset];
			} else {
				&error("[upload_rank] There is no $GameType[$Mode] in $StageName[$Stage].", $errorURL);
			}
		} elsif($Mode == 2) {		# Jet Technique Mode
			my $val = (($Stage * (4 * 5)) + ($Mode * 5));
			if($LevelName[$val] eq '1st') {
				$sName = $StageName[$Stage];
				$lName = "";
				$gName = $GameType[$Mode];
				$text = "Points";
				$offset = 0;
				$value = $rank_point[(($Stage * 4) + 2)][$offset];
				$rank = $rank_rank[(($Stage * 4) + 2)][$offset];
				$char = $rank_player[(($Stage * 4) + 2)][$offset];
			} else {
				&error("[upload_rank] There is no $GameType[$Mode] in $StageName[$Stage].", $errorURL);
			}
		} elsif($Mode == 3) {	# Jet Graffiti Mode
			my $val = (($Stage * (4 * 5)) + ($Mode * 5));
			if($LevelName[$val] eq '1st') {
				$sName = $StageName[$Stage];
				$lName = "";
				$gName = $GameType[$Mode];
				$text = "Points";
				$offset = 0;
				$value = $rank_point[(($Stage * 4) + 3)][$offset];
				$rank = $rank_rank[(($Stage * 4) + 3)][$offset];
				$char = $rank_player[(($Stage * 4) + 3)][$offset];
			} else {
				&error("[upload_rank] There is no $GameType[$Mode] in $StageName[$Stage].", $errorURL);
			}
		}
		else
		{
			&error("[upload_rank] $Mode is an invalid Mode Number.", $errorURL);
		}

		# Output the Ranking Header
		print &html_head("html");

		print <<EOF;
<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.0 Transitional//EN">

<html>
<head>
	<title>Jet Grind Radio - World Rankings Upload</title>
<meta name=x-uirequest content=nohscroll>
</head>

<body bgcolor=#E6CE00 x-marginleft=0 TEXT="#660000" LINK="#B96800" VLINK="#B96800" ALINK="#B96800" marginheight=0 marginwidth=0 topmargin=0 leftmargin=0>



<center>
<TABLE WIDTH="550" BORDER=0 CELLPADDING=0 CELLSPACING=0>
	<tr>
		<TD valign=top width=550><img src="/images/title_rankings.gif" width=550 height=46 border=0 alt=""></td>
	</tr>
</table>
<TABLE WIDTH="550" BORDER=0 CELLPADDING=0 CELLSPACING=0>
	<tr>
		<TD width=7 valign=top><img src="/images/clear.gif" width=7 height=1 border=0 alt=""></td>
		<TD width=368 valign=top>
<FORM METHOD="POST" ACTION="http://jetgrindradio.web.dreamcast.com/cgi-bin/jetgrindradio/score.cgi"> 
<INPUT TYPE="hidden" NAME="mailid" VALUE="$userid"> 
<INPUT TYPE="hidden" NAME="bindata" VALUE="$BINARY">
<INPUT TYPE="hidden" NAME="STAGE" VALUE="$Stage">
<INPUT TYPE="hidden" NAME="MODE" VALUE="$Mode">
<CENTER>
<img src="$StageTbl[$Stage]->[0]" width=320 height=30 border=0 alt=""><BR>
<img src="$StageTbl[$Stage]->[1]" width=320 height=120 border=0 alt=""><BR></CENTER>
			<table border="0" cellpadding=5 cellspacing=0 width=300>
				<tr>
					<td width=100 align=right>Mode:</td>
					<td width=200>$gName $lName</td>
				</tr>
				<tr>
					<td width=100 align=right>$text:</td>
					<td width=200>$value</td>
				</tr>
				<tr>
					<td width=100 align=right>Rating:</td>
					<td width=200><img src="${RankTbl[$rank]->[1]}" border=0></td>
				</tr>
				<tr>
					<td width=100 align=right>Character:</td>
					<td width=200><img src="${CharactersTbl[$char]->[1]}" border=0></td>
				</tr>
			</table>
<br><br>
		
<INPUT TYPE="Image" src="/images/button_continue.gif" width=200 height=30 BORDER="1" NAME="Continue" align=right><br><br><br>

</form>
<a href="JAVASCRIPT:history.back();"><img src="/images/button_back.gif" width=200 height=30 border=1 alt="" align=right></a><br><br></td>
		<TD width=175 valign=top><img src="/images/graphic_rankings.gif" width=175 height=400 border=0 alt=""></td>
	</tr>
</table>
</body>

</html>
EOF
	}
	else
	{
		&error("[upload] Please select a VMU file by clicking on the \"Browse\" button first.", "world_rankings.html");
	}


exit (0);

############################################################################
sub html_head {
    my $type = shift;
    return "Content-type: text/".$type."\n\n";
}

#- Retrieve Cookies From ENV --------------------------------------------------#
sub getCookies {
	# cookies are seperated by a semicolon and a space, this will split
	# them and return a hash of cookies
	my @rawCookies = split (/; /,$ENV{'HTTP_COOKIE'});
	my %cookies;

	foreach(@rawCookies){
	    my ($key, $val) = split (/=/,$_);
	    $cookies{$key} = $val;
	} 

	return %cookies; 
} 
#------------------------------------------------------------------------------#
############################################################################
# error(<error message>, <URL to return>)
# Error Message Display
#
# Argument: Error Message, and URL to return to
############################################################################
sub error {
    my $mes = shift;
	my $URL = shift;

	unless (defined($URL)) {
		$URL = "index.html";
	}

    print <<EOF;
Content-type: text/html

<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.0 Transitional//EN">

<html>
<head>
	<title>Jet Grind Radio - Error</title>
<meta name=x-uirequest content=nohscroll>
</head>

<body bgcolor=#E6CE00 x-marginleft=0 TEXT="#660000" LINK="#B96800" VLINK="#B96800" ALINK="#B96800" marginheight=0 marginwidth=0 topmargin=0 leftmargin=0>



<center>
<TABLE WIDTH="550" BORDER=0 CELLPADDING=0 CELLSPACING=0>
	<tr>
		<TD valign=top width=550><img src="/images/title_error.gif" width=550 height=46 border=0 alt=""></td>
	</tr>
</table>
<TABLE WIDTH="550" BORDER=0 CELLPADDING=0 CELLSPACING=0>
	<tr>
		<TD width=7 valign=top><img src="/images/clear.gif" width=7 height=1 border=0 alt=""></td>
		<TD width=268 valign=top><br>
<b>--ERROR--</b><br><br>

$mes

<br><br><br><br>

<a href="http://jetgrindradio.web.dreamcast.com/$URL"><img src="/images/button_back.gif" width=200 height=30 border=1 alt="" align=right></a></td>
		<TD width=275 valign=top><img src="/images/graphic_login.gif" width=275 height=410 border=0 alt=""></td>
	</tr>
</table>
</body>

</html>
EOF
    exit;
}

#############################################################################
# decode_base64(<base 64'd data stream>)
# returns an array derived from base64 decoding the data stream.
#
# Argument: Character lines encoded by BASE64.
# Back: Encoded bit character lines.
############################################################################
sub decode_base64 {
    my $src = shift;
 
    my @bindata = ();
    my $bytes =0;
    my $tmpstr = "";
    my $bits ="";
    foreach my $c (split(//,$src)) {
        if($c eq '=') {
            last;
        }
        if($c =~ /[A-Z]/) {
            $bits .= substr(unpack("B*",chr(ord($c)-ord('A'))),-6,6);
        } elsif($c =~ /[a-z]/) {
            $bits .= substr(unpack("B*",chr(ord($c)-ord('a')+26)),-6,6);
        } elsif($c =~ /[0-9]/) {
            $bits .= substr(unpack("B*",chr(ord($c)-ord('0')+52)),-6,6);
        } elsif($c  eq '+') {
            $bits .= substr(unpack("B*",chr(62)),-6,6);
        } elsif($c  eq '/') {
            $bits .= substr(unpack("B*",chr(63)),-6,6);
        }

		## now stuff it per-byte into @bindata
        if(length($bits) >= 8) {
            $bindata[$bytes++] = ord(pack("B*",substr($bits,0,8)));
            $bits = substr($bits,8,length($bits)-8);
        }
    }

    return @bindata;
}

# ------------------- Not Used ------------------
#############################################################################
# encode_base64(<raw data stream>)
# returns an array derived from base64 decoding the data stream.
#
# Argument: Raw binary data.
# Back: Encoded bit character lines.
############################################################################
sub encode_base64 {
    my $src = shift;
 
    my $retstr="";
    my $val=0;
    my $bitsnum=0;
    my $tmpval;

    foreach my $c (split(//,$src)) {
        $bitsnum+=8;
        $val=$val*256+pack("c",$c);
        while($bitsnum>=6) {
            $tmpval=($val>>($bitsnum-6))&63;
            $val=$val & ((1<<$bitsnum-6)-1);
            if($tmpval>=0 && $tmpval<26) {
                $retstr .= chr(ord('A')+$tmpval);
            } elsif($tmpval>=26 && $tmpval<52) {
                $retstr .= chr(ord('a')+$tmpval-26);
            } elsif($tmpval>=52 && $tmpval<62) {
                $retstr .= chr(ord('0')+$tmpval-52);
            } elsif($tmpval==62) {
                $retstr .= '+';
            } elsif($tmpval==63) {
                $retstr .= '/';
            }
            if((length($retstr)%73)==72) {
                $retstr .= "\n";
            }
        }
    }
    if($bitsnum>0) {
        $tmpval=($val<<(6-$bitsnum))&63;
        if($tmpval>=0 && $tmpval<26) {
            $retstr .= chr(ord('A')+$tmpval);
        } elsif($tmpval>=26 && $tmpval<52) {
            $retstr .= chr(ord('a')+$tmpval-26);
        } elsif($tmpval>=52 && $tmpval<62) {
            $retstr .= chr(ord('0')+$tmpval-52);
        } elsif($tmpval==62) {
            $retstr .= '+';
        } elsif($tmpval==63) {
            $retstr .= '/';
        }
        if((length($retstr)%73)==72) {
            $retstr .= "\n";
        }
    }
    $retstr .= "==\n";

    return $retstr;
}

#############################################################################
# decode_dp_base64(<DP base 64'd data stream>)
# returns an array derived from DP base64 decoding the data stream.
#
# Argument: Character lines encoded by DP_BASE64.
# Back: Encoded bit character lines.
############################################################################
sub decode_dp_base64 {
	my $src = shift;

	my $bindata = ();
	my $i;

	my $count = 0;
	foreach my $c(split(//,$src))
	{
		if ($c eq '-') {
			last;
		}

		for ($i = 0; $i < 65; $i++)
		{
			if ($c eq $dp_base64[$i]) {
				$count++;
 				$bindata .= $base64[$i];
				if ($count == 77) {
					$count = 0;
					$bindata .= "\n";
				}
				last;
			}
		}
	}

	return $bindata;
}

#############################################################################
# read_multipart_post()
# Read the multipart post request and split into descrete blocks
#
# Argument: None.
# Back: Each component of the Multi-part type.
#############################################################################
sub read_multipart_post {
	my ($formData, $size, @boundry, @returnData, %out, $tmpVar);

    if($ENV{'REQUEST_METHOD'} eq "GET") {
        $formData = $ENV{'QUERY_STRING'};
    }
    elsif($ENV{'REQUEST_METHOD'} eq "POST") {
        $size = $ENV{'CONTENT_LENGTH'};
        read(STDIN, $formData, $size);
    }
    else
    {
        &error("[Upload] Problem figuring out GET or POST\n", "home.html");
    }

    @boundry = split(/\s/, $formData);

    @returnData = split(/$boundry[0]([-]{0,2})/, $formData);
    foreach $tmpVar (@returnData) {
        my $junk = "";
        my $key = "";
        my $value = "";

		if ($tmpVar ne "")
		{
			if ($tmpVar =~ /Content-Disposition: form-data; name=\"(.*)\"/)
			{
				$key = $1;
				($junk, $value) = split(/\".*\"/, $tmpVar, 2);
			}

			if ($key ne "")
			{
				$out{$key} = $value;
			}
		}
	}

    return %out;
}

#############################################################################
# read_vmu_upload()
# Read the VMU Upload Post request, and decode the VMU data
#
# Argument: None.
# Back: Returns an array to the VMU data.
#############################################################################
sub read_vmu_upload {
    my @BINDATA = ();
	my ($key, $value, $tmpVar, @vmuData, %vmuInfo, $filename, $filesize);

	# Get the Post or Get request
	my %toPrint = &read_multipart_post();

	while(($key, $value) = each %toPrint)
	{
		if ($key eq "STAGE")
		{
			# Read the stage the user selected
			$Stage = $value;
		}

		if ($key eq "MODE")
		{
			# Read the stage the user selected
			$Mode = $value;
		}

		if ($key eq "VMUDATA")
		{
			@vmuData = split(/&/, $value);

			foreach $tmpVar (@vmuData) {
				($key, $value) = split(/=/, $tmpVar, 2);
				$key =~ s/%([A-Fa-f0-9]{2})/pack("C",hex($1))/ge;
				$key =~ s/[\n\r]//g;
				$value =~ s/%([A-Fa-f0-9]{2})/pack("C",hex($1))/ge;
				$vmuInfo{$key} = $value;
			}

			while(($key, $value) = each %vmuInfo)
			{
				# get file name
				if ($key eq "filename") {
					$filename = $value;
				}

				# get file size
				if ($key eq "bl") {
					$filesize = $value;
				}

				# get binary data
				if ($key eq "tm")
				{
					my ($trim, $base64);
					$trim = substr($value, 15);
					$base64 = "";
					$base64 = &decode_dp_base64($trim);

					# Save intermediate for later use
					$BINARY = $base64;

					@BINDATA = &decode_base64($base64);
				}
			}
		}
	}

	# No file was uploaded so purge any junk data that was decoded
	if ($filesize == 0)
	{
		@BINDATA = ();
	}

    return @BINDATA;
}


############################################################################
# bindump(<binary data to dump>)
# Will dump out the data as an HTML file
############################################################################
sub bindump {
	my $binary = shift;
	my $max = $#{$binary} + 1;

	print &html_head("html");
	print "<HTML><HEAD><TITLE>Binary Dump</TITLE></HEAD><BODY>\n";
	print "<P>File Size:".$max."</P>\n";

	print "<TABLE BORDER=\"1\">\n";
	for (my $loop = 0; $loop < $max; $loop += 8)
	{
		print "<TR ALIGN=\"CENTER\" VALIGN=\"TOP\">";
		printf "<TD>%04X</TD>", $loop;
		printf "<TD>%02X</TD>", $binary->[$loop+0] & 0xFF;
		printf "<TD>%02X</TD>", $binary->[$loop+1] & 0xFF;
		printf "<TD>%02X</TD>", $binary->[$loop+2] & 0xFF;
		printf "<TD>%02X</TD>", $binary->[$loop+3] & 0xFF;
		printf "<TD>%02X</TD>", $binary->[$loop+4] & 0xFF;
		printf "<TD>%02X</TD>", $binary->[$loop+5] & 0xFF;
		printf "<TD>%02X</TD>", $binary->[$loop+6] & 0xFF;
		printf "<TD>%02X</TD>", $binary->[$loop+7] & 0xFF;
		print "</TR>\n";
	}
	print "</TABLE>\n";
	print "</BODY></HTML>\n";

	exit(0);
}

############################################################################
# make_vmidata(<filename on server>, <filesize in bytes>, <filename on VMU>)
# Creating VMI data.
# returns an array derived from DP base64 decoding the data stream.
#
# Argument: Character lines encoded by DP_BASE64.
# Back: Encoded bit character lines.
############################################################################
sub make_vmidata {
	my ($filename, $filesize, $vmufname) = @_;

	my ($i, @var, $FILE, @vmiheader);
	my @VMIDATA = ();

    @vmiheader = (
         "V", "M", "U", " ", "D", "O", "W", "N", "L", "O", "A", "D", " ", "F", "I", "L",
         "E","\0","\0","\0","\0","\0","\0","\0","\0","\0","\0","\0","\0","\0","\0","\0",
         "S", "E", "G", "A", " ", "E", "N", "T", "E", "R", "P", "R", "I", "S", "E", "S",
        "\0","\0","\0","\0","\0","\0","\0","\0","\0","\0","\0","\0","\0","\0","\0","\0"
        );

    for($i=0;$i<64;$i++) {
        $VMIDATA[$i+4] = ord($vmiheader[$i]);
    }

    my ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst) = localtime(time);
    $mon++;
    $VMIDATA[68]= ($year+1900 & 255);	# year (Low Byte)
    $VMIDATA[69]= ($year+1900 >> 8);	# year (High Byte)
    $VMIDATA[70]= $mon;			# month
    $VMIDATA[71]= $mday;		# day
    $VMIDATA[72]= $hour;		# hour
    $VMIDATA[73]= $min;			# minute
    $VMIDATA[74]= $sec;			# second
    $VMIDATA[75]= $wday;		# day of the week
    $VMIDATA[76]= 1;			# Version (superior rank)
    $VMIDATA[77]= 0;			# Version (inferior rank)
    $VMIDATA[78]= 1;			# # of files (inferior rank)
    $VMIDATA[79]= 0;			# # of files (superior rank)

    @var=split(//,$filename);
    for($i=0;$i<8;$i++) {
        $VMIDATA[$i+80] = ord($var[$i]);	# File names on server
    }

	$VMIDATA[0] = $VMIDATA[80] & 0x53;		# Checksum
	$VMIDATA[1] = $VMIDATA[81] & 0x45;		#
	$VMIDATA[2] = $VMIDATA[82] & 0x47;		#
	$VMIDATA[3] = $VMIDATA[83] & 0x41;		#

	$FILE = $vmufname;
    $FILE =~ y/a-z/A-Z/;

    @var=split(//,$FILE);
    for($i=0;$i<12;$i++) {
        $VMIDATA[$i+88] = ord($var[$i]);	# VMS file names
    }

    $VMIDATA[100] = 0;			# Attributes
    $VMIDATA[101] = 0;			# Attributes
    $VMIDATA[102] = 0;			# Attributes
    $VMIDATA[103] = 0;			# Attributes
    $VMIDATA[104] = 0;			# File size
    $VMIDATA[105] = $filesize*2;	# File size
    $VMIDATA[106] = 0;			# File size
    $VMIDATA[107] = 0;			# File size

	my $vmidata = "";
	for($i=0;$i<108;$i++) {
        $vmidata .= pack("C",$VMIDATA[$i]);
    }

    return $vmidata;
}


#####################################################################
# BinaryToFloat()
# Input : An array of bytes that represent a floating point number.
# Output : The decimal value which is equivalent of the input float.
#####################################################################
sub BinaryToFloat()
{
        my $BinArr = shift;

        my($b0, $b1, $b2, $b3, $sign, $exponent, $non_exponent, $mult, $fraction, $pattern, $numerator, $bitset, @FractionBaseTable, $float, $x, $y, $z, $loop);

        my $str;

# Build the FractionBaseTable.
# It contains decreasing powers of 2 from 22 to 0.
        @FractionBaseTable = ();
        $FractionBaseTable[0] = 4194304;
        $FractionBaseTable[1] = 2097152;
        $FractionBaseTable[2] = 1048576;
        $FractionBaseTable[3] = 524288;
        $FractionBaseTable[4] = 262144;
        $FractionBaseTable[5] = 131072;
        $FractionBaseTable[6] = 65536;
        $FractionBaseTable[7] = 32768;
        $FractionBaseTable[8] = 16384;
        $FractionBaseTable[9] = 8192;
        $FractionBaseTable[10] = 4096;
        $FractionBaseTable[11] = 2048;
        $FractionBaseTable[12] = 1024;
        $FractionBaseTable[13] = 512;
        $FractionBaseTable[14] = 256;
        $FractionBaseTable[15] = 128;
        $FractionBaseTable[16] = 64;
        $FractionBaseTable[17] = 32;
        $FractionBaseTable[18] = 16;
        $FractionBaseTable[19] = 8;
        $FractionBaseTable[20] = 4;
        $FractionBaseTable[21] = 2;
        $FractionBaseTable[22] = 1;
        $FractionBaseTable[23] = 0;

#Read the 4 bytes
        $b0 = $BinArr->[0];
        $b1 = $BinArr->[1];
        $b2 = $BinArr->[2];
        $b3 = $BinArr->[3];

		# Check to see if everything is clear, then output zero
		if (($b0|$b1|$b2|$b3) == 0)	{
			return 0.0;
		}

# Take the sign bit : 7th bit of b3
        $sign = $b3 & 0x80;

# Calculate the exponent now.
        $x = ($b3 << 8) | $b2;
        $exponent = ($x >> 7) & 0xFF;
        $mult = $exponent - 127;

#Calculate the non-exponent component
        $y = ($b2 << 16) | ($b1 << 8) | $b0;
        $z = $y & 0x7FFFFF;
        $fraction = $z << 1;
        $pattern = $fraction;
        $numerator = 0;
        for ($loop = 0; $loop < 23; $loop++)    {
                $bitset = ($pattern & 0x800000);
                if ($bitset)    {
                        $numerator += $FractionBaseTable[$loop];
                }
                $pattern = $pattern << 1;
        }
        $non_exponent = ($numerator / 8388608) + 1.0;
        $float = ($non_exponent) * 2 ** $mult;
        if ($sign)      {
                $float *= (-1);
        }
        return $float;
}

#####################################################################
# CalcTime()
# Input : Time value
# Output : Minutes, seconds, miliseconds
#####################################################################
sub CalcTime()	{
	my $time = shift;

	my ($min, $sec, $msec, $str);

	$min = int($time / 60);
	$sec = $time - ($min * 60);
	$str = sprintf("%d:%02.3f", $min, $sec);

	return $str;
}
 

#####################################################################
# crcCalculateBuffer()
# Input : 
# Output : 
#####################################################################
sub crcCalculateBuffer()
{
	my $byteOffset = shift;
	my $nByteCount = shift;
	my $crcValue = shift;
	my $cBufferPtr = shift;
	my ($crc, $crcPolynomial, $crcData, $i, $j, @crcTable);

	$crcPolynomial = 0xEDB88320;
	@crcTable = ();
	for ($i = 0; $i < 256; $i++) {
		$crc = $i;
		for ($j = 0; $j < 8; $j++) {
			if ($crc & 0x0001) {
				$crc = ($crc >> 1) ^ $crcPolynomial;
			}
			else {
				$crc = $crc >> 1;
			}
		}
		$crcTable[$i] = $crc;
	}
	#Calculate the new CRC here :-
	$crc = $crcValue;
	for ($i = 0; $i < $nByteCount; $i++) {
		$crcData = ($crc ^ $cBufferPtr->[$byteOffset + $i]) & 0xFF;
		$crc = (($crc >> 8) & 0x00FFFFFF) ^ $crcTable[$crcData];
	}
	return $crc;
}

#############################################################################
# decode_jgrData(<offset>, <binary data>)
# returns an binary array derived from JGR encoded the data stream.
#
# Argument: Binary data encoded by JGR.
# Back: Unencoded binary data.
############################################################################
sub decode_jgrData {
	my $offset = shift;
    my $bindata = shift;
	
	my @dk = ();
	my ($i, $j, $shift, $nk, $size);
	
	
	# Defined Values
	my $SAVELOAD_KEY_NUM = 8;
	my $SYSTEM_SAVE_DATA = 1008;

	# Size of the actual ranking data
	$size = ($SYSTEM_SAVE_DATA - $SAVELOAD_KEY_NUM);

	# Read the last eight bytes
	for ($i = 0; $i < $SAVELOAD_KEY_NUM; $i++)
	{
		$dk[$i] = $bindata->[$offset + $size + $i];
					
		if ($dk[$i] & 0x80)
		{
			# Turn it into a signed number
			$dk[$i] = sprintf "%d", $dk[$i] |= 0xFFFFFF00;
			$dk[$i] = unpack('s', pack('S', $dk[$i]));
		}

		if ($dk[$i] == 0)
		{
			# An Error condition occurred.
			$bindata->[$offset + 0] = 0;
			$bindata->[$offset + 1] = 0;
			$bindata->[$offset + 2] = 0;
			$bindata->[$offset + 3] = 0;

			return;
		}
	}
	
	
	# Decode the binary data
	for ($i = $size - 1; $i >= 0; $i--)
	{
		$j = $i % $SAVELOAD_KEY_NUM;		
		# 3bit Decoding
		$nk = ($dk[$j] >> 3) & 0x1f;		
		$nk |= ($dk[$j] << 5) & 0xe0;
		
		
		if ($nk & 0x80)
		{
			# Turn it into a signed number
			$nk = sprintf ("%d", $nk |= 0xFFFFFF00);	
			$nk = unpack('s', pack('S', $nk));
		}
						
		$dk[$j] = $nk;
		
		$shift = $i + $dk[$j];
		#print $i.": ".$nk.", ".$shift.";";
		
		if ($shift < 0) {
			$shift += $size;
		}
		elsif ($shift >= $size) {
			$shift -= $size;
		}
		$nk = $bindata->[$offset + $i];
		$bindata->[$offset + $i] = $bindata->[$offset + $shift];
		$bindata->[$offset + $shift] = $nk ^ $bindata->[$offset + $i];
		

	}
	
	

    return;
}

